#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    ConcreteModel,
    Var,
    value,
    Constraint,
    Expression,
    Objective,
    TransformationFactory,
    assert_optimal_termination,
    units as pyunits,
    NonNegativeReals,
)
from pyomo.network import Arc

from idaes.core import FlowsheetBlock, UnitModelCostingBlock
from watertap.core.solvers import get_solver
from idaes.core.util.initialization import propagate_state
import idaes.core.util.model_statistics as mstat
from idaes.models.unit_models import Feed, Product, Separator, Mixer
from watertap.unit_models.pressure_changer import Pump
from idaes.models.unit_models.mixer import MixingType, MomentumMixingType
import pandas as pd
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslogger
from watertap.unit_models.electrodialysis_1D import (
    ElectricalOperationMode,
    PressureDropMethod,
    FrictionFactorMethod,
    HydraulicDiameterMethod,
    LimitingCurrentDensityMethod,
)
from watertap.unit_models.electrodialysis_1D import Electrodialysis1D
from watertap.costing.watertap_costing_package import WaterTAPCosting
from watertap.property_models.multicomp_aq_sol_prop_pack import MCASParameterBlock

__author__ = "Xiangyu Bi"
_log = idaeslogger.getLogger(__name__)

"""
This flowsheet implements a single-stack electrodialysis desaliantion system in a 
feed-and-bleed mode, which recirculate a part of concentrate solution to achieve a
desired water reovery > 50%. 
"""


def main():
    m = build()
    solver = get_solver()

    ## Simulate a fully defined operation
    # A fully-defined system simulation, dof=0
    init_arg = {
        ("flow_vol_phase", ("Liq")): 5.2e-4,
        ("conc_mol_phase_comp", ("Liq", "Na_+")): 34.188,
        ("conc_mol_phase_comp", ("Liq", "Cl_-")): 34.188,
    }  # Corresponding to C_feed = 2g/L
    m.fs.feed.properties.calculate_state(
        init_arg,
        hold_state=True,
    )
    m.fs.EDstack.voltage_applied[0].fix(10)
    m.fs.recovery_vol_H2O.fix(0.7)
    _condition_base(m)

    # Initialize and solve the model
    initialize_system(m, solver=solver)
    solve(m, solver=solver)
    print("\n***---Fully-defined simulation results, Fixed inlet and voltage---***")
    display_model_metrics(m)

    # Perform an optimization over cell_length, cell_pair_mum, and voltage_applied

    ed = m.fs.EDstack
    ulim = (
        ed.voltage_x[0, 0].value
        / ed.current_density_x[0, 0].value
        * ed.current_dens_lim_x[0, 0].value
        * 1.5
    )
    m.fs.EDstack.voltage_applied[0].unfix()
    m.fs.EDstack.voltage_applied[0].setlb(0.1)
    m.fs.EDstack.voltage_applied[0].setub(ulim)
    m.fs.EDstack.cell_pair_num.unfix()
    m.fs.EDstack.cell_pair_num.setlb(10)
    m.fs.EDstack.cell_pair_num.setub(1000)
    m.fs.EDstack.cell_length.unfix()
    m.fs.prod.properties[0].conc_mol_phase_comp["Liq", "Na_+"].fix(
        1.7094
    )  # Corresponding to C_product = 100 ppm
    m.fs.objective = Objective(expr=m.fs.costing.LCOW)
    solve(m, solver=solver, tee=True)
    m.fs.EDstack.cell_pair_num.fix(round(m.fs.EDstack.cell_pair_num.value))
    solve(m, solver=solver, tee=True)
    print("\n***---Optimization results, Product conc = 100 ppb---***")
    display_model_metrics(m)

    return m


def build():
    # ---building model---
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    ion_dict = {
        "solute_list": ["Na_+", "Cl_-"],
        "mw_data": {"H2O": 18e-3, "Na_+": 23e-3, "Cl_-": 35.5e-3},
        "elec_mobility_data": {("Liq", "Na_+"): 5.19e-8, ("Liq", "Cl_-"): 7.92e-8},
        "diffusivity_data": {("Liq", "Na_+"): 1.33e-9, ("Liq", "Cl_-"): 2.03e-9},
        "charge": {"Na_+": 1, "Cl_-": -1},
    }
    m.fs.properties = MCASParameterBlock(**ion_dict)
    m.fs.costing = WaterTAPCosting()
    m.fs.feed = Feed(property_package=m.fs.properties)
    m.fs.sepa0 = Separator(
        property_package=m.fs.properties,
        outlet_list=["to_dil_in", "to_conc_in0"],
    )
    m.fs.mix0 = Mixer(
        property_package=m.fs.properties,
        energy_mixing_type=MixingType.none,
        momentum_mixing_type=MomentumMixingType.none,
        inlet_list=["from_feed", "from_conc_out"],
    )

    m.fs.pump0 = Pump(property_package=m.fs.properties)
    m.fs.pump0.ratioP_calculation.deactivate()
    m.fs.pump1 = Pump(property_package=m.fs.properties)
    m.fs.pump0.ratioP_calculation.deactivate()

    # Add electrodialysis (ED) stacks
    m.fs.EDstack = Electrodialysis1D(
        property_package=m.fs.properties,
        operation_mode=ElectricalOperationMode.Constant_Voltage,
        finite_elements=20,
        has_pressure_change=True,
        has_nonohmic_potential_membrane=True,
        has_Nernst_diffusion_layer=True,
        limiting_current_density_method=LimitingCurrentDensityMethod.Theoretical,
        pressure_drop_method=PressureDropMethod.Darcy_Weisbach,
        hydraulic_diameter_method=HydraulicDiameterMethod.spacer_specific_area_known,
        friction_factor_method=FrictionFactorMethod.Gurreri,
    )
    m.fs.sepa1 = Separator(
        property_package=m.fs.properties,
        outlet_list=["to_disp", "to_conc_in1"],
    )
    m.fs.prod = Product(property_package=m.fs.properties)
    m.fs.disp = Product(property_package=m.fs.properties)

    # touch necessary variables to ensure they are constructed and solved
    m.fs.prod.properties[0].flow_vol_phase[...]
    m.fs.disp.properties[0].flow_vol_phase[...]

    # costing
    m.fs.EDstack.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method_arguments={"cost_electricity_flow": True, "has_rectifier": True},
    )
    m.fs.pump0.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.pump1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.costing.cost_process()
    m.fs.costing.add_annual_water_production(
        m.fs.prod.properties[0].flow_vol_phase["Liq"]
    )
    m.fs.costing.add_LCOW(m.fs.prod.properties[0].flow_vol)
    m.fs.costing.add_specific_energy_consumption(
        m.fs.prod.properties[0].flow_vol_phase["Liq"]
    )

    # add extra variables and constraints
    m.fs.recovery_vol_H2O = Var(
        initialize=0.5,
        bounds=(0, 1),
        domain=NonNegativeReals,
        units=pyunits.dimensionless,
        doc="flowsheet level water recovery calculated by volumeric flow rate",
    )

    m.fs.feed_salinity = Var(
        initialize=1,
        domain=NonNegativeReals,
        units=pyunits.kg * pyunits.m**-3,
        doc="Salinity of feed solution. Tested only with NaCl.",
    )

    m.fs.eq_recovery_vol_H2O = Constraint(
        expr=m.fs.recovery_vol_H2O * m.fs.feed.properties[0].flow_vol_phase["Liq"]
        == m.fs.prod.properties[0].flow_vol_phase["Liq"]
    )

    m.fs.eq_electrodialysis_equal_flow = Constraint(
        expr=m.fs.EDstack.diluate.properties[0, 0].flow_vol_phase["Liq"]
        == m.fs.EDstack.concentrate.properties[0, 0].flow_vol_phase["Liq"]
    )
    m.fs.eq_feed_salinity = Constraint(
        expr=m.fs.feed_salinity
        == sum(
            m.fs.feed.properties[0].conc_mass_phase_comp["Liq", j]
            for j in m.fs.properties.ion_set
        )
    )
    m.fs.product_salinity = Expression(
        expr=sum(
            m.fs.prod.properties[0].conc_mass_phase_comp["Liq", j]
            for j in m.fs.properties.solute_set
        )
    )
    m.fs.disposal_salinity = Expression(
        expr=sum(
            m.fs.disp.properties[0].conc_mass_phase_comp["Liq", j]
            for j in m.fs.properties.solute_set
        )
    )

    m.fs.mem_area = Expression(
        expr=m.fs.EDstack.cell_width
        * m.fs.EDstack.cell_length
        * m.fs.EDstack.cell_pair_num
    )

    m.fs.voltage_per_cp = Expression(
        expr=m.fs.EDstack.voltage_applied[0] / m.fs.EDstack.cell_pair_num
    )

    # Add Arcs
    m.fs.arc0 = Arc(source=m.fs.feed.outlet, destination=m.fs.sepa0.inlet)
    m.fs.arc1b = Arc(source=m.fs.sepa0.to_dil_in, destination=m.fs.pump1.inlet)
    m.fs.arc1f = Arc(source=m.fs.pump1.outlet, destination=m.fs.EDstack.inlet_diluate)
    m.fs.arc2 = Arc(
        source=m.fs.sepa0.to_conc_in0,
        destination=m.fs.mix0.from_feed,
    )
    m.fs.arc3b = Arc(source=m.fs.mix0.outlet, destination=m.fs.pump0.inlet)
    m.fs.arc3f = Arc(
        source=m.fs.pump0.outlet, destination=m.fs.EDstack.inlet_concentrate
    )
    m.fs.arc4 = Arc(source=m.fs.EDstack.outlet_diluate, destination=m.fs.prod.inlet)
    m.fs.arc5 = Arc(
        source=m.fs.EDstack.outlet_concentrate, destination=m.fs.sepa1.inlet
    )
    m.fs.arc6 = Arc(source=m.fs.sepa1.to_disp, destination=m.fs.disp.inlet)
    m.fs.arc7 = Arc(source=m.fs.sepa1.to_conc_in1, destination=m.fs.mix0.from_conc_out)
    TransformationFactory("network.expand_arcs").apply_to(m)

    return m


def _condition_base(m):
    # ---specifications---
    # Here is simulated a scenario of a defined EDstack and
    # specific water recovery and product salinity.
    m.fs.feed.properties[0].pressure.fix(101325)
    m.fs.feed.properties[0].temperature.fix(298.15)
    m.fs.pump0.control_volume.properties_in[0].pressure.fix(101325)
    m.fs.pump1.efficiency_pump.fix(0.8)
    m.fs.pump0.efficiency_pump.fix(0.8)

    m.fs.prod.properties[0].pressure.fix(101325)
    m.fs.disp.properties[0].pressure.fix(101325)
    m.fs.disp.properties[0].temperature.fix(298.15)

    # Set ED unit vars
    # membrane properties
    m.fs.EDstack.water_trans_number_membrane["cem"].fix(5.8)
    m.fs.EDstack.water_trans_number_membrane["aem"].fix(4.3)
    m.fs.EDstack.water_permeability_membrane["cem"].fix(2.16e-14)
    m.fs.EDstack.water_permeability_membrane["aem"].fix(1.75e-14)
    m.fs.EDstack.membrane_areal_resistance["cem"].fix(1.89e-4)
    m.fs.EDstack.membrane_areal_resistance["aem"].fix(1.77e-4)
    m.fs.EDstack.membrane_areal_resistance_coef["cem"].fix(0)
    m.fs.EDstack.membrane_areal_resistance_coef["aem"].fix(0)
    m.fs.EDstack.solute_diffusivity_membrane["cem", "Na_+"].fix(3.28e-11)
    m.fs.EDstack.solute_diffusivity_membrane["aem", "Na_+"].fix(3.28e-11)
    m.fs.EDstack.solute_diffusivity_membrane["cem", "Cl_-"].fix(3.28e-11)
    m.fs.EDstack.solute_diffusivity_membrane["aem", "Cl_-"].fix(3.28e-11)
    m.fs.EDstack.ion_trans_number_membrane["cem", "Na_+"].fix(1)
    m.fs.EDstack.ion_trans_number_membrane["aem", "Na_+"].fix(0)
    m.fs.EDstack.ion_trans_number_membrane["cem", "Cl_-"].fix(0)
    m.fs.EDstack.ion_trans_number_membrane["aem", "Cl_-"].fix(1)
    m.fs.EDstack.membrane_thickness["aem"].fix(1.3e-4)
    m.fs.EDstack.membrane_thickness["cem"].fix(1.3e-4)

    # Stack properties
    m.fs.EDstack.cell_pair_num.fix(56)
    m.fs.EDstack.channel_height.fix(7.1e-4)
    m.fs.EDstack.cell_width.fix(0.197)
    m.fs.EDstack.cell_length.fix(1.68)

    # Spacer properties
    m.fs.EDstack.spacer_porosity.fix(0.83)
    m.fs.EDstack.spacer_specific_area.fix(10400)
    m.fs.EDstack.spacer_conductivity_coefficient.fix(1)

    # Electrochemical properties
    m.fs.EDstack.electrodes_resistance.fix(0)
    m.fs.EDstack.current_utilization.fix(1)
    m.fs.EDstack.diffus_mass.fix(1.6e-9)

    assert mstat.degrees_of_freedom(m) == 0


def solve(m, solver=None, tee=True, check_termination=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(m, tee=tee)
    if check_termination:
        assert_optimal_termination(results)
    return results


def initialize_system(m, solver=None):
    # set up solver
    if solver is None:
        solver = get_solver()
    optarg = solver.options
    # set scaling factors for state vars and call the 'calculate_scaling_factors' function
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 0.1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e2, index=("Liq", "Na_+")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e2, index=("Liq", "Cl_-")
    )
    iscale.set_scaling_factor(m.fs.EDstack.cell_width, 5)
    iscale.set_scaling_factor(m.fs.EDstack.cell_length, 1)
    iscale.set_scaling_factor(m.fs.EDstack.cell_pair_num, 0.1)
    iscale.set_scaling_factor(m.fs.EDstack.voltage_applied, 1)
    iscale.set_scaling_factor(m.fs.pump0.control_volume.work, 1e1)
    iscale.set_scaling_factor(m.fs.pump1.control_volume.work, 1e1)
    iscale.calculate_scaling_factors(m)
    iscale.constraint_scaling_transform(
        m.fs.eq_recovery_vol_H2O,
        10 * iscale.get_scaling_factor(m.fs.feed.properties[0].flow_vol_phase["Liq"]),
    )
    iscale.constraint_scaling_transform(
        m.fs.eq_electrodialysis_equal_flow,
        10 * iscale.get_scaling_factor(m.fs.feed.properties[0].flow_vol_phase["Liq"]),
    )
    iscale.constraint_scaling_transform(
        m.fs.eq_feed_salinity,
        iscale.get_scaling_factor(
            m.fs.feed.properties[0].conc_mass_phase_comp["Liq", "Na_+"]
        ),
    )

    # populate intitial properties throughout the system
    m.fs.feed.initialize(optarg=optarg)
    propagate_state(m.fs.arc0)
    m.fs.sepa0.initialize(optarg=optarg)
    propagate_state(m.fs.arc1b)
    m.fs.pump1.deltaP[0].fix(2e5)
    m.fs.pump1.initialize()
    m.fs.pump1.deltaP[0].unfix()
    propagate_state(destination=m.fs.pump0.inlet, source=m.fs.pump1.inlet)
    m.fs.pump0.deltaP[0].fix(2e5)
    m.fs.pump0.initialize()
    m.fs.pump0.deltaP[0].unfix()
    propagate_state(m.fs.arc1f)
    propagate_state(m.fs.arc3f)
    m.fs.EDstack.initialize(optarg=optarg)
    propagate_state(m.fs.arc4)
    m.fs.prod.initialize(optarg=optarg)
    propagate_state(m.fs.arc5)
    m.fs.prod.initialize(optarg=optarg)
    m.fs.sepa1.split_fraction[0, "to_conc_in1"].fix(
        max(2 - value(m.fs.recovery_vol_H2O) ** -1, 1e-8)
    )
    m.fs.sepa1.initialize(optarg=optarg)
    m.fs.sepa1.split_fraction[0, "to_conc_in1"].unfix()
    propagate_state(m.fs.arc6)
    m.fs.disp.initialize()
    propagate_state(m.fs.arc7)
    propagate_state(m.fs.arc3b, direction="backward")
    propagate_state(m.fs.arc2)
    m.fs.mix0.initialize(optarg=optarg)
    m.fs.costing.initialize()

    iscale.calculate_scaling_factors(m)


def display_model_metrics(m):

    print("---Report the ED unit---")

    m.fs.EDstack.report()

    print("---Flow properties in feed, product and disposal---")

    fp = {
        "Feed": [
            value(m.fs.feed.properties[0].flow_vol_phase["Liq"]),
            value(
                sum(
                    m.fs.feed.properties[0].conc_mass_phase_comp["Liq", j]
                    for j in m.fs.properties.ion_set
                )
            ),
        ],
        "Product": [
            value(m.fs.prod.properties[0].flow_vol_phase["Liq"]),
            value(m.fs.product_salinity),
        ],
        "Disposal": [
            value(m.fs.disp.properties[0].flow_vol_phase["Liq"]),
            value(m.fs.disposal_salinity),
        ],
    }
    fp_table = pd.DataFrame(
        data=fp,
        index=["Volume Flow Rate (m3/s)", "Total Ion Mass Concentration (kg/m3)"],
    )
    print(fp_table)

    print("---Performance Metrics---")

    pm_table = pd.DataFrame(
        data=[
            value(m.fs.recovery_vol_H2O),
            value(m.fs.mem_area),
            value(m.fs.EDstack.cell_pair_num),
            value(m.fs.EDstack.cell_length),
            value(m.fs.EDstack.voltage_applied[0]),
            value(m.fs.costing.specific_energy_consumption),
            value(m.fs.costing.LCOW),
        ],
        columns=["value"],
        index=[
            "Water recovery by volume",
            "Total membrane area (aem or cem), m2",
            "ED cell pair number",
            "ED cell flow path length",
            "Operation Voltage, V",
            "Specific energy consumption, kWh/m3",
            "Levelized cost of water, $/m3",
        ],
    )
    print(pm_table)
    print("---Pressure and Temperature point checking---")

    pt_dict = {
        "Feed": (
            value(m.fs.feed.outlet.pressure[0]),
            value(m.fs.feed.outlet.temperature[0]),
        ),
        "ED_in_dil": (
            value(m.fs.EDstack.inlet_diluate.pressure[0]),
            value(m.fs.EDstack.inlet_diluate.temperature[0]),
        ),
        "ED_in_conc": (
            value(m.fs.EDstack.inlet_concentrate.pressure[0]),
            value(m.fs.EDstack.inlet_concentrate.temperature[0]),
        ),
        "ED_out_dil": (
            value(m.fs.EDstack.outlet_diluate.pressure[0]),
            value(m.fs.EDstack.outlet_diluate.temperature[0]),
        ),
        "ED_out_conc": (
            value(m.fs.EDstack.outlet_concentrate.pressure[0]),
            value(m.fs.EDstack.outlet_concentrate.temperature[0]),
        ),
        "Prod": (
            value(m.fs.prod.inlet.pressure[0]),
            value(m.fs.prod.inlet.temperature[0]),
        ),
        "Disp": (
            value(m.fs.disp.inlet.pressure[0]),
            value(m.fs.disp.inlet.temperature[0]),
        ),
    }
    pt_table = pd.DataFrame(data=pt_dict, index=["Pressure (Pa)", "Temperature (K)"])
    print(pt_table)


if __name__ == "__main__":
    m = main()



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Thermophysical property package to be used in conjunction with ASM1 reactions.
"""

# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    StateBlockData,
    StateBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    LiquidPhase,
    Component,
    Solute,
    Solvent,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import fix_state_vars, revert_state_vars
from idaes.core.scaling import CustomScalerBase
import idaes.logger as idaeslog
from idaes.core.base.property_base import PhysicalParameterBlock
import idaes.core.util.scaling as iscale

# Some more information about this module
__author__ = "Andrew Lee, Adam Atia, Xinhong Liu"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("ASM1ParameterBlock")
class ASM1ParameterData(PhysicalParameterBlock):
    """
    Property Parameter Block Class
    """

    def build(self):
        """
        Callable method for Block construction.
        """
        super().build()

        self._state_block_class = ASM1StateBlock

        # Add Phase objects
        self.Liq = LiquidPhase()

        # Add Component objects
        self.H2O = Solvent()

        self.S_I = Solute(doc="Soluble inert organic matter, S_I")
        self.S_S = Solute(doc="Readily biodegradable substrate, S_S")
        self.X_I = Solute(doc="Particulate inert organic matter, X_I")
        self.X_S = Solute(doc="Slowly biodegradable substrate, X_S")
        self.X_BH = Solute(doc="Active heterotrophic biomass, X_B,H")
        self.X_BA = Solute(doc="Active autotrophic biomass, X_B,A")
        self.X_P = Solute(doc="Particulate products arising from biomass decay, X_P")
        self.S_O = Solute(doc="Oxygen, S_O")
        self.S_NO = Solute(doc="Nitrate and nitrite nitrogen, S_NO")
        self.S_NH = Solute(doc="NH4+ + NH3 nitrogen, S_NH")
        self.S_ND = Solute(doc="Soluble biodegradable organic nitrogen, S_ND")
        self.X_ND = Solute(doc="Particulate biodegradable organic nitrogen, X_ND")

        self.S_ALK = Component(doc="Alkalinity, S_ALK")

        # Create sets for use across ASM models and associated unit models (e.g., thickener, dewaterer)
        self.non_particulate_component_set = pyo.Set(
            initialize=[
                "S_I",
                "S_S",
                "S_O",
                "S_NO",
                "S_NH",
                "S_ND",
                "H2O",
                "S_ALK",
            ]
        )
        self.particulate_component_set = pyo.Set(
            initialize=["X_I", "X_S", "X_P", "X_BH", "X_BA", "X_ND"]
        )
        self.tss_component_set = pyo.Set(
            initialize=["X_I", "X_S", "X_P", "X_BH", "X_BA"]
        )

        # Heat capacity of water
        self.cp_mass = pyo.Param(
            initialize=4182,
            doc="Specific heat capacity of water",
            units=pyo.units.J / pyo.units.kg / pyo.units.K,
        )
        # Density of water
        self.dens_mass = pyo.Param(
            initialize=997,
            doc="Density of water",
            units=pyo.units.kg / pyo.units.m**3,
        )

        # Thermodynamic reference state
        self.pressure_ref = pyo.Param(
            within=pyo.PositiveReals,
            mutable=True,
            default=101325.0,
            doc="Reference pressure",
            units=pyo.units.Pa,
        )
        self.temperature_ref = pyo.Param(
            within=pyo.PositiveReals,
            mutable=True,
            default=298.15,
            doc="Reference temperature",
            units=pyo.units.K,
        )
        self.f_p = pyo.Var(
            initialize=0.08,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Fraction of biomass yielding particulate products, f_p",
        )
        self.i_xb = pyo.Var(
            initialize=0.08,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Mass fraction of N per COD in biomass, i_xb",
        )
        self.i_xp = pyo.Var(
            initialize=0.06,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Mass fraction of N per COD in particulates, i_xp",
        )
        self.COD_to_SS = pyo.Var(
            initialize=0.75,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Conversion factor applied for TSS calculation",
        )
        self.BOD5_factor = pyo.Param(
            ["raw", "effluent"],
            initialize={"raw": 0.65, "effluent": 0.25},
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Conversion factor for BOD5",
        )
        # Fix Vars that are treated as Params
        for v in self.component_objects(pyo.Var):
            v.fix()

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties(
            {
                "flow_vol": {"method": None},
                "pressure": {"method": None},
                "temperature": {"method": None},
                "conc_mass_comp": {"method": None},
            }
        )
        obj.define_custom_properties(
            {
                "alkalinity": {"method": None},
                "TSS": {"method": "_TSS"},
                "BOD5": {"method": "_BOD5"},
                "TKN": {"method": "_TKN"},
                "Total_N": {"method": "_Total_N"},
                "COD": {"method": "_COD"},
            }
        )
        obj.add_default_units(
            {
                "time": pyo.units.s,
                "length": pyo.units.m,
                "mass": pyo.units.kg,
                "amount": pyo.units.kmol,
                "temperature": pyo.units.K,
            }
        )


class ASM1PropertiesScaler(CustomScalerBase):
    """
    Scaler for the Activated Sludge Model No.1 property package.

    Flow and temperature are scaled by the default value (if no user input provided), and
    pressure is scaled assuming an order of magnitude of 1e5 Pa.
    """

    UNIT_SCALING_FACTORS = {
        # "QuantityName: (reference units, scaling factor)
        "Pressure": (pyo.units.Pa, 1e-6),
    }

    DEFAULT_SCALING_FACTORS = {
        "flow_vol": 1e1,
        "temperature": 1e-1,
    }

    def variable_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        self.scale_variable_by_default(model.temperature, overwrite=overwrite)
        self.scale_variable_by_default(model.flow_vol, overwrite=overwrite)
        self.scale_variable_by_units(model.pressure, overwrite=overwrite)

    # There are currently no constraints in this model
    def constraint_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        pass


class _ASM1StateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    default_scaler = ASM1PropertiesScaler

    def initialize(
        self,
        state_args=None,
        state_vars_fixed=False,
        hold_state=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for property package.

        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provided at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.The keys for the state_args dictionary are:
            flow_vol : value at which to initialize total volumetric flow (default=None)
            alkalinity: value of alkalinity expressed as molar concentration
            conc_mass_comp : value at which to initialize component concentrations (default=None)
            pressure : value at which to initialize pressure (default=None)
            temperature : value at which to initialize temperature (default=None)
            outlvl : sets output level of initialization routine
            state_vars_fixed: Flag to denote if state vars have already been fixed.
                              True - states have already been fixed and
                              initialization does not need to worry
                              about fixing and unfixing variables.
                              False - states have not been fixed. The state
                              block will deal with fixing/unfixing.
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         True - states variables are not unfixed, and
                         a dict of returned containing flags for
                         which states were fixed during initialization.
                         False - state variables are unfixed after
                         initialization by calling the release_state method.

        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")

        if state_vars_fixed is False:
            # Fix state variables if not already fixed
            flags = fix_state_vars(self, state_args)

        else:
            # Check when the state vars are fixed already result in dof 0
            for k in self.keys():
                if degrees_of_freedom(self[k]) != 0:
                    raise Exception(
                        "State vars fixed but degrees of freedom "
                        "for state block is not zero during "
                        "initialization."
                    )

        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                self.release_state(flags)

        init_log.info("Initialization Complete.")

    def release_state(self, flags, outlvl=idaeslog.NOTSET):
        """
        Method to relase state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")

        if flags is None:
            return
        # Unfix state variables
        revert_state_vars(self, flags)
        init_log.info("State Released.")


@declare_process_block_class("ASM1StateBlock", block_class=_ASM1StateBlock)
class ASM1StateBlockData(StateBlockData):
    """
    StateBlock for calculating thermophysical properties associated with the ASM1
    reaction system.
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super().build()

        # Create state variables
        self.flow_vol = pyo.Var(
            initialize=1.0,
            domain=pyo.NonNegativeReals,
            doc="Total volumetric flowrate",
            units=pyo.units.m**3 / pyo.units.s,
        )
        self.pressure = pyo.Var(
            domain=pyo.NonNegativeReals,
            initialize=101325.0,
            bounds=(1e3, 1e6),
            doc="Pressure",
            units=pyo.units.Pa,
        )
        self.temperature = pyo.Var(
            domain=pyo.NonNegativeReals,
            initialize=298.15,
            bounds=(273.15, 323.15),
            doc="Temperature",
            units=pyo.units.K,
        )
        self.conc_mass_comp = pyo.Var(
            self.params.solute_set,
            domain=pyo.NonNegativeReals,
            initialize=0.1,
            doc="Component mass concentrations",
            units=pyo.units.kg / pyo.units.m**3,
        )
        self.alkalinity = pyo.Var(
            domain=pyo.NonNegativeReals,
            initialize=1,
            doc="Alkalinity in molar concentration",
            units=pyo.units.kmol / pyo.units.m**3,
        )

        # Material and energy flow and density expressions
        def material_flow_expression(self, j):
            if j == "H2O":
                return self.flow_vol * self.params.dens_mass
            elif j == "S_ALK":
                # Convert moles of alkalinity to mass of C assuming all is HCO3-
                return (
                    self.flow_vol
                    * self.alkalinity
                    * (12 * pyo.units.kg / pyo.units.kmol)
                )
            else:
                return self.flow_vol * self.conc_mass_comp[j]

        self.material_flow_expression = pyo.Expression(
            self.component_list,
            rule=material_flow_expression,
            doc="Material flow terms",
        )

        def enthalpy_flow_expression(self):
            return (
                self.flow_vol
                * self.params.dens_mass
                * self.params.cp_mass
                * (self.temperature - self.params.temperature_ref)
            )

        self.enthalpy_flow_expression = pyo.Expression(
            rule=enthalpy_flow_expression, doc="Enthalpy flow term"
        )

        def material_density_expression(self, j):
            if j == "H2O":
                return self.params.dens_mass
            elif j == "S_ALK":
                # Convert moles of alkalinity to mass of C assuming all is HCO3-
                return self.alkalinity * (12 * pyo.units.kg / pyo.units.kmol)
            else:
                return self.conc_mass_comp[j]

        self.material_density_expression = pyo.Expression(
            self.component_list,
            rule=material_density_expression,
            doc="Material density terms",
        )

        def energy_density_expression(self):
            return (
                self.params.dens_mass
                * self.params.cp_mass
                * (self.temperature - self.params.temperature_ref)
            )

        self.energy_density_expression = pyo.Expression(
            rule=energy_density_expression, doc="Energy density term"
        )

        def _TSS(self):
            tss = (
                self.conc_mass_comp["X_S"]
                + self.conc_mass_comp["X_I"]
                + self.conc_mass_comp["X_BH"]
                + self.conc_mass_comp["X_BA"]
                + self.conc_mass_comp["X_P"]
            )
            return self.params.COD_to_SS * tss

        self.TSS = pyo.Expression(
            rule=_TSS,
            doc="Total suspended solids (TSS)",
        )

        def _BOD5(self, i):
            bod5 = (
                self.conc_mass_comp["X_S"]
                + self.conc_mass_comp["X_S"]
                + (1 - self.params.f_p)
                * (self.conc_mass_comp["X_BH"] + self.conc_mass_comp["X_BA"])
            )
            # TODO: 0.25 should be a parameter instead as it changes by influent/effluent
            return self.params.BOD5_factor[i] * bod5

        self.BOD5 = pyo.Expression(
            ["raw", "effluent"],
            rule=_BOD5,
            doc="Five-day Biological Oxygen Demand (BOD5)",
        )

        def _COD(self):
            cod = (
                self.conc_mass_comp["S_S"]
                + self.conc_mass_comp["S_I"]
                + self.conc_mass_comp["X_S"]
                + self.conc_mass_comp["X_S"]
                + self.conc_mass_comp["X_I"]
                + self.conc_mass_comp["X_BH"]
                + self.conc_mass_comp["X_BA"]
                + self.conc_mass_comp["X_P"]
            )
            return cod

        self.COD = pyo.Expression(
            rule=_COD,
            doc="Chemical Oxygen Demand",
        )

        def _TKN(self):
            tkn = (
                self.conc_mass_comp["S_NH"]
                + self.conc_mass_comp["S_ND"]
                + self.conc_mass_comp["X_ND"]
                + self.params.i_xb
                * (self.conc_mass_comp["X_BH"] + self.conc_mass_comp["X_BA"])
                + self.params.i_xp
                * (self.conc_mass_comp["X_P"] + self.conc_mass_comp["X_I"])
            )
            return tkn

        self.TKN = pyo.Expression(
            rule=_TKN,
            doc="Total Kjeldahl Nitrogen",
        )

        def _Total_N(self):
            totaln = self.TKN + self.conc_mass_comp["S_NO"]
            return totaln

        self.Total_N = pyo.Expression(
            rule=_Total_N,
            doc="Total Nitrogen",
        )

    def get_material_flow_terms(self, p, j):
        return self.material_flow_expression[j]

    def get_enthalpy_flow_terms(self, p):
        return self.enthalpy_flow_expression

    def get_material_density_terms(self, p, j):
        return self.material_density_expression[j]

    def get_energy_density_terms(self, p):
        return self.energy_density_expression

    def default_material_balance_type(self):
        return MaterialBalanceType.componentPhase

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def define_state_vars(self):
        return {
            "flow_vol": self.flow_vol,
            "alkalinity": self.alkalinity,
            "conc_mass_comp": self.conc_mass_comp,
            "temperature": self.temperature,
            "pressure": self.pressure,
        }

    def define_display_vars(self):
        return {
            "Volumetric Flowrate": self.flow_vol,
            "Molar Alkalinity": self.alkalinity,
            "Mass Concentration": self.conc_mass_comp,
            "Temperature": self.temperature,
            "Pressure": self.pressure,
        }

    def get_material_flow_basis(self):
        return MaterialFlowBasis.mass

    def calculate_scaling_factors(self):
        # Get default scale factors and do calculations from base classes
        super().calculate_scaling_factors()

        # No constraints in this model as yet, just need to set scaling factors
        # for expressions
        sf_F = iscale.get_scaling_factor(self.flow_vol, default=1e2, warning=True)
        sf_T = iscale.get_scaling_factor(self.temperature, default=1e-2, warning=True)

        # Mass flow and density terms
        for j in self.component_list:
            if j == "H2O":
                sf_C = pyo.value(1 / self.params.dens_mass)
            elif j == "S_ALK":
                sf_C = 1e-1 * iscale.get_scaling_factor(
                    self.alkalinity, default=1, warning=True
                )
            else:
                sf_C = iscale.get_scaling_factor(
                    self.conc_mass_comp[j], default=1e2, warning=True
                )

            iscale.set_scaling_factor(self.material_flow_expression[j], sf_F * sf_C)
            iscale.set_scaling_factor(self.material_density_expression[j], sf_C)

        # Enthalpy and energy terms
        sf_rho_cp = pyo.value(1 / (self.params.dens_mass * self.params.cp_mass))
        iscale.set_scaling_factor(
            self.enthalpy_flow_expression, sf_F * sf_rho_cp * sf_T
        )
        iscale.set_scaling_factor(self.energy_density_expression, sf_rho_cp * sf_T)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from watertap.property_models.unit_specific.coagulation_prop_pack import (
    CoagulationParameterBlock,
)
from watertap.property_models.NaCl_prop_pack import NaClParameterBlock
from watertap.property_models.seawater_prop_pack import SeawaterParameterBlock
from watertap.unit_models.coag_floc_model import CoagulationFlocculation
from pyomo.environ import (
    ConcreteModel,
    value,
    units as pyunits,
)
from idaes.core import FlowsheetBlock
from idaes.core.util.exceptions import ConfigurationError
import idaes.core.util.scaling as iscale
from watertap.core.solvers import get_solver
from watertap.unit_models.tests.unit_test_harness import UnitTestHarness
import re

__author__ = "Austin Ladshaw"

solver = get_solver()


# -----------------------------------------------------------------------------
def build():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = CoagulationParameterBlock()
    ## NOTE: These values provided are just DUMMY values for the purposes
    #        of testing. They are not meant to be representative of any
    #        particular chemicals or real-world additives.
    chem_dict = {
        "Alum": {
            "parameter_data": {
                "mw_additive": (200, pyunits.g / pyunits.mol),
                "moles_salt_per_mole_additive": 3,
                "mw_salt": (100, pyunits.g / pyunits.mol),
            }
        },
        "Poly": {
            "parameter_data": {
                "mw_additive": (25, pyunits.g / pyunits.mol),
                "moles_salt_per_mole_additive": 0,
                "mw_salt": (23, pyunits.g / pyunits.mol),
            }
        },
    }
    m.fs.unit = CoagulationFlocculation(
        property_package=m.fs.properties, chemical_additives=chem_dict
    )

    m.fs.unit.fix_tss_turbidity_relation_defaults()
    m.fs.unit.initial_turbidity_ntu.fix()
    m.fs.unit.final_turbidity_ntu.fix(5)
    m.fs.unit.chemical_doses[0, "Alum"].fix(10)
    m.fs.unit.chemical_doses[0, "Poly"].fix(5)

    # set the inlet streams
    m.fs.unit.inlet.pressure.fix(101325)
    m.fs.unit.inlet.temperature.fix(298.15)
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(1)
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(0.01)
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TSS"].fix(0.01)
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "Sludge"].fix(0.0)

    # set performance vars
    m.fs.unit.rapid_mixing_retention_time[0].fix(60)
    m.fs.unit.num_rapid_mixing_basins.fix(4)
    m.fs.unit.rapid_mixing_vel_grad[0].fix(750)

    m.fs.unit.floc_retention_time[0].fix(1800)
    m.fs.unit.single_paddle_length.fix(4)
    m.fs.unit.single_paddle_width.fix(0.5)
    m.fs.unit.paddle_rotational_speed[0].fix(0.03)

    m.fs.unit.paddle_drag_coef[0].fix(1.5)
    m.fs.unit.vel_fraction.fix(0.7)
    m.fs.unit.num_paddle_wheels.fix(4)
    m.fs.unit.num_paddles_per_wheel.fix(4)

    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_in[0.0].mass_frac_phase_comp[
            "Liq", "Sludge"
        ],
        1e12,
    )
    iscale.calculate_scaling_factors(m.fs.unit)

    return m


def build_no_chemicals():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = CoagulationParameterBlock()
    m.fs.unit = CoagulationFlocculation(property_package=m.fs.properties)

    m.fs.unit.fix_tss_turbidity_relation_defaults()
    m.fs.unit.initial_turbidity_ntu.fix(5000)
    m.fs.unit.final_turbidity_ntu.fix(100)

    tss_in = value(m.fs.unit.compute_inlet_tss_mass_flow(0))

    m.fs.unit.inlet.pressure.fix(101325)
    m.fs.unit.inlet.temperature.fix(298.15)
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(1)
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(0.01)
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TSS"].fix(tss_in)
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "Sludge"].fix(0.0)

    # set performance vars
    m.fs.unit.rapid_mixing_retention_time[0].fix(60)
    m.fs.unit.num_rapid_mixing_basins.fix(4)
    m.fs.unit.rapid_mixing_vel_grad[0].fix(750)

    m.fs.unit.floc_retention_time[0].fix(1800)
    m.fs.unit.single_paddle_length.fix(4)
    m.fs.unit.single_paddle_width.fix(0.5)
    m.fs.unit.paddle_rotational_speed[0].fix(0.03)

    m.fs.unit.paddle_drag_coef[0].fix(1.5)
    m.fs.unit.vel_fraction.fix(0.7)
    m.fs.unit.num_paddle_wheels.fix(4)
    m.fs.unit.num_paddles_per_wheel.fix(4)

    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_in[0.0].mass_frac_phase_comp[
            "Liq", "Sludge"
        ],
        1e12,
    )
    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestCoagFloc(UnitTestHarness):
    def configure(self):
        m = build()

        self.unit_solutions[m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]] = 1
        self.unit_solutions[
            m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "Sludge"]
        ] = 0.009990636
        self.unit_solutions[m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "TDS"]] = (
            0.01001510
        )
        self.unit_solutions[m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "TSS"]] = (
            9.36352627e-6
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TDS"]
                + m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TSS"]
                + m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "Sludge"],
                "out": m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "TDS"]
                + m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "TSS"]
                + m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "Sludge"],
            },
        }

        return m


class TestCoagFlocNoChemicals(UnitTestHarness):
    def configure(self):
        m = build_no_chemicals()

        self.unit_solutions[m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]] = 1
        self.unit_solutions[
            m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "Sludge"]
        ] = 0.009112749
        self.unit_solutions[m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "TDS"]] = (
            0.01
        )
        self.unit_solutions[m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "TSS"]] = (
            0.0001872506
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TDS"]
                + m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TSS"]
                + m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "Sludge"],
                "out": m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "TDS"]
                + m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "TSS"]
                + m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "Sludge"],
            },
        }

        return m


class TestCoagFlocErrorLog:
    @pytest.mark.unit
    def test_dictionary_error(self):

        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = CoagulationParameterBlock()

        bad_dict1 = {
            "Alum": {
                "foo_bar": {
                    "mw_additive": (200, pyunits.g / pyunits.mol),
                    "moles_salt_per_mole_additive": 3,
                    "mw_salt": (100, pyunits.g / pyunits.mol),
                }
            }
        }

        with pytest.raises(
            ConfigurationError,
            match="Did not provide a 'parameter_data' for chemical",
        ):
            m.fs.unit = CoagulationFlocculation(
                property_package=m.fs.properties, chemical_additives=bad_dict1
            )

        bad_dict2 = {
            "Alum": {
                "parameter_data": {
                    "foo_bar": (200, pyunits.g / pyunits.mol),
                    "moles_salt_per_mole_additive": 3,
                    "mw_salt": (100, pyunits.g / pyunits.mol),
                }
            }
        }
        with pytest.raises(
            ConfigurationError, match="Did not provide a 'mw_additive' for chemical"
        ):
            m.fs.unit = CoagulationFlocculation(
                property_package=m.fs.properties, chemical_additives=bad_dict2
            )

        bad_dict3 = {
            "Alum": {
                "parameter_data": {
                    "mw_additive": (200, pyunits.g / pyunits.mol),
                    "moles_salt_per_mole_additive": "foo-bar",
                    "mw_salt": (100, pyunits.g / pyunits.mol),
                }
            }
        }
        with pytest.raises(
            ConfigurationError,
            match="Did not provide a number for 'moles_salt_per_mole_additive'",
        ):
            m.fs.unit = CoagulationFlocculation(
                property_package=m.fs.properties, chemical_additives=bad_dict3
            )

        bad_dict4 = {
            "Alum": {
                "parameter_data": {
                    "mw_additive": (200, pyunits.g / pyunits.mol),
                    "foo-bar": 3,
                    "mw_salt": (100, pyunits.g / pyunits.mol),
                }
            }
        }
        with pytest.raises(
            ConfigurationError,
            match="Did not provide a 'moles_salt_per_mole_additive' for chemical",
        ):
            m.fs.unit = CoagulationFlocculation(
                property_package=m.fs.properties, chemical_additives=bad_dict4
            )

        bad_dict5 = {
            "Alum": {
                "parameter_data": {
                    "mw_additive": (200, pyunits.g / pyunits.mol),
                    "moles_salt_per_mole_additive": 3,
                    "foo-bar": (100, pyunits.g / pyunits.mol),
                }
            }
        }
        with pytest.raises(
            ConfigurationError, match="Did not provide a 'mw_salt' for chemical"
        ):
            m.fs.unit = CoagulationFlocculation(
                property_package=m.fs.properties, chemical_additives=bad_dict5
            )

        bad_dict6 = {
            "Alum": {
                "parameter_data": {
                    "mw_additive": "not a tuple",
                    "moles_salt_per_mole_additive": 3,
                    "mw_salt": (100, pyunits.g / pyunits.mol),
                }
            }
        }
        with pytest.raises(
            ConfigurationError, match="Did not provide a tuple for 'mw_additive'"
        ):
            m.fs.unit = CoagulationFlocculation(
                property_package=m.fs.properties, chemical_additives=bad_dict6
            )

    @pytest.mark.unit
    def test_property_error(self):

        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        error_msg = (
            "Coagulation-Flocculation model MUST contain ('Liq','TDS') "
            "as a component, but the property package has only specified "
            "the following components [('Liq', 'H2O'), ('Liq', 'NaCl')]"
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.properties = NaClParameterBlock()
            m.fs.unit = CoagulationFlocculation(property_package=m.fs.properties)

        error_msg = (
            "Coagulation-Flocculation model MUST contain ('Liq','Sludge') "
            "as a component, but the property package has only specified "
            "the following components [('Liq', 'H2O'), ('Liq', 'TDS')]"
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            m.fs.properties = SeawaterParameterBlock()
            m.fs.unit = CoagulationFlocculation(property_package=m.fs.properties)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import (
    ConcreteModel,
    value,
    Var,
    Objective,
    assert_optimal_termination,
    Block,
)
from pyomo.network import Port
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from watertap.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.unit_models import Feed, Product
from watertap.unit_models.ion_exchange_0D import (
    IonExchange0D,
    IonExchangeType,
    RegenerantChem,
    IsothermType,
)
from watertap.property_models.multicomp_aq_sol_prop_pack import MCASParameterBlock

import watertap.flowsheets.ion_exchange.ion_exchange_demo as ixf

import math

__author__ = "Kurban Sitterley"


solver = get_solver()


class TestIXDemoCa:

    @pytest.fixture(scope="class")
    def ix_0D_Ca(self):
        target_ion = "Ca_2+"
        ions = [target_ion]
        m = ixf.ix_build(ions)
        return m

    @pytest.mark.unit
    def test_build_model(self, ix_0D_Ca):
        m = ix_0D_Ca

        # Test basic build
        assert isinstance(m, ConcreteModel)
        assert isinstance(m.fs, FlowsheetBlock)
        assert isinstance(m.fs.properties, MCASParameterBlock)
        assert isinstance(m.fs.costing, Block)
        assert isinstance(m.fs.feed, Feed)
        assert isinstance(m.fs.ion_exchange, IonExchange0D)
        assert isinstance(m.fs.product, Product)

        # Test port
        assert isinstance(m.fs.feed.outlet, Port)
        assert isinstance(m.fs.ion_exchange.inlet, Port)
        assert isinstance(m.fs.ion_exchange.outlet, Port)
        assert isinstance(m.fs.product.inlet, Port)

        # # Test consting
        assert isinstance(m.fs.ion_exchange.costing, Block)
        assert isinstance(m.fs.ion_exchange.costing.capital_cost, Var)
        assert isinstance(m.fs.ion_exchange.costing.fixed_operating_cost, Var)

        var_str_list = [
            "total_capital_cost",
            "total_operating_cost",
        ]
        for var_str in var_str_list:
            var = getattr(m.fs.costing, var_str)
            assert isinstance(var, Var)

        # Test arcs
        arc_dict = {
            m.fs.feed_to_ix: (m.fs.feed.outlet, m.fs.ion_exchange.inlet),
            m.fs.ix_to_product: (m.fs.ion_exchange.outlet, m.fs.product.inlet),
        }
        for arc, port_tpl in arc_dict.items():
            assert arc.source is port_tpl[0]
            assert arc.destination is port_tpl[1]

        # test configrations
        assert len(m.fs.ion_exchange.config) == 11
        assert not m.fs.ion_exchange.config.dynamic
        assert not m.fs.ion_exchange.config.has_holdup
        assert m.fs.ion_exchange.config.target_ion == "Ca_2+"
        assert m.fs.ion_exchange.ion_exchange_type == IonExchangeType.cation
        assert m.fs.ion_exchange.config.regenerant == RegenerantChem.NaCl
        assert m.fs.ion_exchange.config.isotherm == IsothermType.langmuir

        assert m.fs.ion_exchange.config.property_package is m.fs.properties
        assert "H2O" in m.fs.properties.component_list

        assert_units_consistent(m)

    @pytest.mark.component
    def test_specific_operating_conditions(self, ix_0D_Ca):

        m = ix_0D_Ca
        ixf.set_operating_conditions(m)
        ixf.initialize_system(m)
        assert degrees_of_freedom(m) == 0

        solver = get_solver()
        results = solver.solve(m)
        assert_optimal_termination(results)
        assert value(m.fs.feed.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            0.05, rel=1e-3
        )
        assert value(m.fs.product.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            0.049995010, rel=1e-3
        )
        assert value(
            sum(
                m.fs.feed.properties[0].conc_mass_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        ) == pytest.approx(0.1, rel=1e-3)
        assert value(
            sum(
                m.fs.product.properties[0].conc_mass_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        ) == pytest.approx(0.000211438, rel=1e-3)

        results_dict = {
            "resin_diam": 0.0007,
            "resin_bulk_dens": 0.7,
            "resin_surf_per_vol": 4285.71,
            "c_norm": {"Ca_2+": 0.473078},
            "bed_vol_tot": 11.999,
            "bed_depth": 1.7,
            "bed_porosity": 0.5,
            "col_height": 3.488715,
            "col_diam": 1.498964,
            "col_height_to_diam_ratio": 2.3274,
            "number_columns": 4.0,
            "t_breakthru": 56759.757,
            "ebct": 240.0,
            "vel_bed": 0.007083333,
            "service_flow_rate": 15.0,
            "N_Re": 4.95833,
            "N_Sc": {"Ca_2+": 1086.956},
            "N_Sh": {"Ca_2+": 26.296},
            "N_Pe_particle": 0.10782,
            "N_Pe_bed": 261.867,
            "resin_max_capacity": 3.0,
            "resin_eq_capacity": 1.6857,
            "resin_unused_capacity": 1.3142,
            "langmuir": {"Ca_2+": 0.7},
            "mass_removed": {"Ca_2+": 7079.969},
            "num_transfer_units": 35.548,
            "dimensionless_time": 1.0,
            "partition_ratio": 235.998,
            "fluid_mass_transfer_coeff": {"Ca_2+": 3.456e-05},
            "pressure_drop": 9.450141,
            "bed_vol": 2.999999,
            "t_rinse": 1200.0,
            "t_waste": 3600.0,
            "t_cycle": 60359.757,
            "regen_pump_power": 0.040480054,
            "regen_tank_vol": 29.999,
            "bw_flow": 0.009803921,
            "bed_expansion_frac": 0.46395,
            "rinse_flow": 0.049999999,
            "bw_pump_power": 0.007937265,
            "rinse_pump_power": 0.080960109,
            "bed_expansion_h": 0.788715,
            "main_pump_power": 3.82939,
            "col_vol_per": 6.156555,
            "col_vol_tot": 24.626,
            "t_contact": 120.0,
            "vel_inter": 0.014166,
            "bv_calc": 236.498,
            "lh": 0.0,
            "separation_factor": {"Ca_2+": 1.4285},
            "rate_coeff": {"Ca_2+": 0.000211597},
            "HTU": {"Ca_2+": 0.047822147},
        }
        for v, r in results_dict.items():
            ixv = getattr(m.fs.ion_exchange, v)
            if ixv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(ixv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(ixv)

        sys_cost_results = {
            "aggregate_capital_cost": 810841.86,
            "aggregate_fixed_operating_cost": 4099.26,
            "aggregate_variable_operating_cost": 0.0,
            "aggregate_flow_electricity": 3.9587,
            "aggregate_flow_NaCl": 2352713.22,
            "aggregate_flow_costs": {"electricity": 2429.18, "NaCl": 214194.76},
            "total_capital_cost": 810841.86,
            "total_operating_cost": 223386.06,
            "aggregate_direct_capital_cost": 405420.93,
            "maintenance_labor_chemical_operating_cost": 24325.25,
            "total_fixed_operating_cost": 28424.51,
            "total_variable_operating_cost": 194961.55,
            "total_annualized_cost": 304470.25,
            "annual_water_production": 1419950.29,
            "LCOW": 0.214423,
            "specific_energy_consumption": 0.021995,
        }

        for v, r in sys_cost_results.items():
            mv = getattr(m.fs.costing, v)
            if mv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(mv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(mv)

    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_optimization(self, ix_0D_Ca):
        m = ix_0D_Ca
        ixf.optimize_system(m)
        isinstance(m.fs.obj, Objective)
        assert m.fs.obj.expr == m.fs.costing.LCOW
        assert degrees_of_freedom(m) == 2

        assert value(m.fs.feed.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            0.05, rel=1e-3
        )
        assert value(m.fs.product.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            0.04999624, rel=1e-3
        )
        assert value(
            sum(
                m.fs.feed.properties[0].conc_mass_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        ) == pytest.approx(0.1, rel=1e-3)
        assert value(
            sum(
                m.fs.product.properties[0].conc_mass_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        ) == pytest.approx(0.025, rel=1e-3)
        ix = m.fs.ion_exchange
        num_col = math.ceil(
            ix.number_columns()
        )  # To eliminate fractional number of columns
        bed_depth = ix.bed_depth()
        ix.bed_depth.fix(bed_depth)
        ix.number_columns.fix(num_col)
        results = solver.solve(m)
        assert_optimal_termination(results)
        assert degrees_of_freedom(m) == 0

        results_dict = {
            "resin_surf_per_vol": 4285.714,
            "c_norm": {"Ca_2+": 0.99398},
            "bed_depth": 2.211,
            "t_breakthru": 133829.3,
            "vel_bed": 0.009213559,
            "N_Re": 6.4494,
            "N_Sc": {"Ca_2+": 1086.956},
            "N_Sh": {"Ca_2+": 28.755},
            "N_Pe_particle": 0.122332,
            "N_Pe_bed": 386.44,
            "resin_eq_capacity": 2.9873,
            "resin_unused_capacity": 0.01266295,
            "mass_removed": {"Ca_2+": 12546.81},
            "num_transfer_units": 38.872,
            "dimensionless_time": 1.3321,
            "partition_ratio": 418.227,
            "fluid_mass_transfer_coeff": {"Ca_2+": 3.7792e-05},
            "pressure_drop": 16.049,
            "bed_vol": 2.399,
            "lh": 12.909,
            "separation_factor": {"Ca_2+": 1.4285},
            "rate_coeff": {"Ca_2+": 0.0002313},
            "HTU": {"Ca_2+": 0.056884},
        }

        for v, r in results_dict.items():
            ixv = getattr(m.fs.ion_exchange, v)
            if ixv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(ixv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(ixv)

        sys_cost_results = {
            "aggregate_capital_cost": 827291.05,
            "aggregate_fixed_operating_cost": 3935.28,
            "aggregate_variable_operating_cost": 0.0,
            "aggregate_flow_electricity": 6.8301,
            "aggregate_flow_NaCl": 991992.09,
            "aggregate_flow_costs": {"electricity": 4191.08, "NaCl": 90312.54},
            "total_capital_cost": 827291.05,
            "total_operating_cost": 113807.28,
            "aggregate_direct_capital_cost": 413645.52,
            "maintenance_labor_chemical_operating_cost": 24818.73,
            "total_fixed_operating_cost": 28754.02,
            "total_variable_operating_cost": 85053.26,
            "total_annualized_cost": 196536.39,
            "annual_water_production": 1419985.49,
            "LCOW": 0.1384073,
            "specific_energy_consumption": 0.0379478,
        }

        for v, r in sys_cost_results.items():
            mv = getattr(m.fs.costing, v)
            if mv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(mv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(mv)

    @pytest.mark.unit
    @pytest.mark.requires_idaes_solver
    def test_main_fun(self):
        m = ixf.main()

        assert degrees_of_freedom(m) == 0
        results_dict = {
            "resin_surf_per_vol": 4285.71,
            "c_norm": {"Ca_2+": 0.993980},
            "bed_vol_tot": 11.9999,
            "bed_depth": 2.2112,
            "bed_porosity": 0.5,
            "col_height": 4.2371,
            "col_diam": 1.1755,
            "number_columns": 5.0,
            "t_breakthru": 133829.3,
            "ebct": 240,
            "vel_bed": 0.00921355996,
            "service_flow_rate": 15.0,
            "N_Re": 6.44949197,
            "N_Sc": {"Ca_2+": 1086.95},
            "N_Sh": {"Ca_2+": 28.755},
            "N_Pe_particle": 0.122332,
            "N_Pe_bed": 386.4407,
            "resin_eq_capacity": 2.9873,
            "langmuir": {"Ca_2+": 0.7},
            "mass_removed": {"Ca_2+": 12546.81},
            "num_transfer_units": 38.8726,
            "dimensionless_time": 1.33210,
            "partition_ratio": 418.23,
            "fluid_mass_transfer_coeff": {"Ca_2+": 3.7792874e-05},
        }

        for v, r in results_dict.items():
            ixv = getattr(m.fs.ion_exchange, v)
            if ixv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(ixv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(ixv)

        sys_cost_results = {
            "aggregate_capital_cost": 827291.06,
            "aggregate_fixed_operating_cost": 3935.28,
            "aggregate_variable_operating_cost": 0.0,
            "aggregate_flow_electricity": 6.8301,
            "aggregate_flow_NaCl": 991992.09,
            "aggregate_flow_costs": {"electricity": 4191.09, "NaCl": 90312.54},
            "total_capital_cost": 827291.06,
            "total_operating_cost": 113807.28,
            "aggregate_direct_capital_cost": 413645.52,
            "maintenance_labor_chemical_operating_cost": 24818.73,
            "total_fixed_operating_cost": 28754.02,
            "total_variable_operating_cost": 85053.26,
            "total_annualized_cost": 196536.39,
            "annual_water_production": 1419985.49,
            "LCOW": 0.1384073,
            "specific_energy_consumption": 0.03794785,
        }
        for v, r in sys_cost_results.items():
            mv = getattr(m.fs.costing, v)
            if mv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(mv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(mv)


class TestIXDemoSO4:

    @pytest.fixture(scope="class")
    def ix_0D_SO4(self):
        target_ion = "SO4_2-"
        ions = [target_ion]
        m = ixf.ix_build(ions)
        return m

    @pytest.mark.unit
    def test_build_model(self, ix_0D_SO4):
        m = ix_0D_SO4

        # Test basic build
        assert isinstance(m, ConcreteModel)
        assert isinstance(m.fs, FlowsheetBlock)
        assert isinstance(m.fs.properties, MCASParameterBlock)
        assert isinstance(m.fs.costing, Block)
        assert isinstance(m.fs.feed, Feed)
        assert isinstance(m.fs.ion_exchange, IonExchange0D)
        assert isinstance(m.fs.product, Product)

        # Test port
        assert isinstance(m.fs.feed.outlet, Port)
        assert isinstance(m.fs.ion_exchange.inlet, Port)
        assert isinstance(m.fs.ion_exchange.outlet, Port)
        assert isinstance(m.fs.product.inlet, Port)

        # # Test consting
        assert isinstance(m.fs.ion_exchange.costing, Block)
        assert isinstance(m.fs.ion_exchange.costing.capital_cost, Var)
        assert isinstance(m.fs.ion_exchange.costing.fixed_operating_cost, Var)

        var_str_list = [
            "total_capital_cost",
            "total_operating_cost",
        ]
        for var_str in var_str_list:
            var = getattr(m.fs.costing, var_str)
            assert isinstance(var, Var)

        # Test arcs
        arc_dict = {
            m.fs.feed_to_ix: (m.fs.feed.outlet, m.fs.ion_exchange.inlet),
            m.fs.ix_to_product: (m.fs.ion_exchange.outlet, m.fs.product.inlet),
        }
        for arc, port_tpl in arc_dict.items():
            assert arc.source is port_tpl[0]
            assert arc.destination is port_tpl[1]

        # test configrations
        assert len(m.fs.ion_exchange.config) == 11
        assert not m.fs.ion_exchange.config.dynamic
        assert not m.fs.ion_exchange.config.has_holdup
        assert m.fs.ion_exchange.config.target_ion == "SO4_2-"
        assert m.fs.ion_exchange.ion_exchange_type == IonExchangeType.anion
        assert m.fs.ion_exchange.config.regenerant == RegenerantChem.NaCl
        assert m.fs.ion_exchange.config.isotherm == IsothermType.langmuir

        assert m.fs.ion_exchange.config.property_package is m.fs.properties
        assert "H2O" in m.fs.properties.component_list

        assert_units_consistent(m)

    @pytest.mark.component
    def test_specific_operating_conditions(self, ix_0D_SO4):

        m = ix_0D_SO4
        ixf.set_operating_conditions(m)
        ixf.initialize_system(m)
        assert degrees_of_freedom(m) == 0

        solver = get_solver()
        results = solver.solve(m)
        assert_optimal_termination(results)
        assert value(m.fs.feed.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            0.05, rel=1e-3
        )
        assert value(m.fs.product.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            0.049995010, rel=1e-3
        )
        assert value(
            sum(
                m.fs.feed.properties[0].conc_mass_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        ) == pytest.approx(0.1, rel=1e-3)
        assert value(
            sum(
                m.fs.product.properties[0].conc_mass_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        ) == pytest.approx(8.821e-5, rel=1e-3)

        results_dict = {
            "resin_surf_per_vol": 4285.714,
            "c_norm": {"SO4_2-": 0.473},
            "bed_vol_tot": 12.0,
            "col_height": 3.4887,
            "col_diam": 1.498,
            "col_height_to_diam_ratio": 2.327,
            "number_columns": 4.0,
            "t_breakthru": 136055.4,
            "ebct": 240.0,
            "vel_bed": 0.0070833,
            "N_Re": 4.958333,
            "N_Sc": {"SO4_2-": 943.3},
            "N_Sh": {"SO4_2-": 25.095},
            "N_Pe_particle": 0.107827,
            "N_Pe_bed": 261.867,
            "resin_max_capacity": 3.0,
            "resin_eq_capacity": 1.685707,
            "resin_unused_capacity": 1.314292,
            "langmuir": {"SO4_2-": 0.7},
            "mass_removed": {"SO4_2-": 7079.969},
            "num_transfer_units": 39.087,
            "dimensionless_time": 1.0,
            "partition_ratio": 566.397,
            "fluid_mass_transfer_coeff": {"SO4_2-": 3.8001e-05},
            "pressure_drop": 9.450141,
            "separation_factor": {"SO4_2-": 1.4285},
            "rate_coeff": {"SO4_2-": 0.000232663},
            "HTU": {"SO4_2-": 0.043492261},
        }
        for v, r in results_dict.items():
            ixv = getattr(m.fs.ion_exchange, v)
            if ixv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(ixv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(ixv)

        sys_cost_results = {
            "aggregate_capital_cost": 866570.3,
            "aggregate_fixed_operating_cost": 5492.468,
            "aggregate_variable_operating_cost": 0.0,
            "aggregate_flow_electricity": 4.023,
            "aggregate_flow_NaCl": 1016854.2,
            "aggregate_flow_costs": {"electricity": 2468.7, "NaCl": 92576.0},
            "total_capital_cost": 866570.3,
            "total_operating_cost": 117029.8,
            "aggregate_direct_capital_cost": 433285.1,
            "maintenance_labor_chemical_operating_cost": 25997.1,
            "total_fixed_operating_cost": 31489.578,
            "total_variable_operating_cost": 85540.2,
            "total_annualized_cost": 203686.8,
            "annual_water_production": 1419950.1,
            "LCOW": 0.14344,
            "specific_energy_consumption": 0.022353,
        }

        for v, r in sys_cost_results.items():
            mv = getattr(m.fs.costing, v)
            if mv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(mv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(mv)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    ConcreteModel,
    value,
    Constraint,
    Objective,
    Var,
    NonNegativeReals,
    TransformationFactory,
    units as pyunits,
    check_optimal_termination,
)
from pyomo.network import Arc
from idaes.core import FlowsheetBlock
from watertap.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import (
    propagate_state,
)
from idaes.models.unit_models import Heater, Separator, Mixer, Product, Feed
from idaes.models.unit_models.separator import SplittingType
from idaes.models.unit_models.mixer import MomentumMixingType
from idaes.models.unit_models.heat_exchanger import (
    HeatExchanger,
    HeatExchangerFlowPattern,
)
from watertap.unit_models.mvc.components.lmtd_chen_callback import (
    delta_temperature_chen_callback,
)
from watertap.unit_models.pressure_changer import Pump
from idaes.core import UnitModelCostingBlock, FlowDirection
import idaes.core.util.scaling as iscale
from watertap.unit_models.MD.membrane_distillation_0D import MembraneDistillation0D
from watertap.unit_models.MD.MD_channel_base import (
    ConcentrationPolarizationType,
    TemperaturePolarizationType,
    MassTransferCoefficient,
    PressureChangeType,
)
import watertap.property_models.seawater_prop_pack as props_sw
import watertap.property_models.water_prop_pack as props_w
from watertap.costing import WaterTAPCosting
from watertap.costing.unit_models.pump import (
    cost_pump,
)
from watertap.costing.unit_models.heater_chiller import (
    cost_heater_chiller,
)
from watertap.core.util.initialization import (
    assert_degrees_of_freedom,
    interval_initializer,
)

__author__ = "Elmira Shamlou"


def main():
    solver = get_solver()
    m = build()
    set_operating_conditions(m)
    initialize_system(m, solver=solver)

    optimize_set_up(m)

    interval_initializer(m)
    solve(m, solver=solver)

    print("\n***---optimization results---***")
    display_system(m)
    display_design(m)
    display_state(m)

    return m


def build():

    # flowsheet set up
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # properties
    m.fs.properties_hot_ch = props_sw.SeawaterParameterBlock()
    m.fs.properties_cold_ch = props_sw.SeawaterParameterBlock()
    m.fs.properties_vapor = props_w.WaterParameterBlock()

    # Control volume flow blocks
    m.fs.feed = Feed(property_package=m.fs.properties_hot_ch)
    m.fs.permeate = Product(property_package=m.fs.properties_cold_ch)
    m.fs.reject = Product(property_package=m.fs.properties_hot_ch)

    # overall process water recovery
    m.fs.overall_recovery = Var(
        initialize=0.5,
        bounds=(0, 1),
        domain=NonNegativeReals,
        units=pyunits.dimensionless,
        doc="System Water Recovery",
    )
    m.fs.eq_water_recovery = Constraint(
        expr=m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"]
        * m.fs.overall_recovery
        == m.fs.permeate.properties[0].flow_mass_phase_comp["Liq", "H2O"]
    )

    # unit models

    m.fs.MD = MembraneDistillation0D(
        hot_ch={
            "property_package": m.fs.properties_hot_ch,
            "property_package_vapor": m.fs.properties_vapor,
            "has_pressure_change": True,
            "temperature_polarization_type": TemperaturePolarizationType.calculated,
            "concentration_polarization_type": ConcentrationPolarizationType.calculated,
            "mass_transfer_coefficient": MassTransferCoefficient.calculated,
            "pressure_change_type": PressureChangeType.calculated,
            "flow_direction": FlowDirection.forward,
        },
        cold_ch={
            "property_package": m.fs.properties_cold_ch,
            "property_package_vapor": m.fs.properties_vapor,
            "has_pressure_change": True,
            "temperature_polarization_type": TemperaturePolarizationType.calculated,
            "mass_transfer_coefficient": MassTransferCoefficient.none,
            "concentration_polarization_type": ConcentrationPolarizationType.none,
            "pressure_change_type": PressureChangeType.calculated,
            "flow_direction": FlowDirection.backward,
        },
    )

    m.fs.hx = HeatExchanger(
        hot_side_name="hot",
        cold_side_name="cold",
        hot={"property_package": m.fs.properties_hot_ch, "has_pressure_change": True},
        cold={"property_package": m.fs.properties_cold_ch, "has_pressure_change": True},
        delta_temperature_callback=delta_temperature_chen_callback,
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
    )
    # Set lower bound of approach temperatures
    m.fs.hx.delta_temperature_in.setlb(0)
    m.fs.hx.delta_temperature_out.setlb(0)
    m.fs.hx.area.setlb(10)

    m.fs.heater = Heater(
        property_package=m.fs.properties_hot_ch, has_pressure_change=True
    )
    m.fs.chiller = Heater(
        property_package=m.fs.properties_cold_ch, has_pressure_change=True
    )

    m.fs.eq_equal_flow = Constraint(
        expr=m.fs.chiller.control_volume.properties_out[0].flow_mass_phase_comp[
            "Liq", "H2O"
        ]
        == m.fs.heater.control_volume.properties_out[0].flow_mass_phase_comp[
            "Liq", "H2O"
        ]
    )

    m.fs.separator_permeate = Separator(
        property_package=m.fs.properties_cold_ch,
        outlet_list=["permeate", "cold_loop_stream"],
        split_basis=SplittingType.totalFlow,
    )
    m.fs.separator_concentrate = Separator(
        property_package=m.fs.properties_cold_ch,
        outlet_list=["reject", "recycle"],
        split_basis=SplittingType.totalFlow,
    )

    m.fs.recycle_ratio = Var(
        m.fs.config.time,
        initialize=10,
        bounds=(0, 100),
        domain=NonNegativeReals,
        units=pyunits.dimensionless,
        doc="System Water Recovery",
    )
    m.fs.eq_recycle_ratio = Constraint(
        expr=sum(m.fs.feed.flow_mass_phase_comp[0, "Liq", j] for j in ["H2O", "TDS"])
        * m.fs.recycle_ratio[0]
        == sum(
            m.fs.separator_concentrate.recycle.flow_mass_phase_comp[0, "Liq", j]
            for j in ["H2O", "TDS"]
        )
    )

    m.fs.mixer = Mixer(
        property_package=m.fs.properties_hot_ch,
        momentum_mixing_type=MomentumMixingType.equality,
        inlet_list=["feed", "recycle"],
    )
    m.fs.mixer.pressure_equality_constraints[0, 2].deactivate()

    m.fs.pump_feed = Pump(property_package=m.fs.properties_hot_ch)
    m.fs.pump_permeate = Pump(property_package=m.fs.properties_cold_ch)
    m.fs.pump_brine = Pump(property_package=m.fs.properties_hot_ch)

    # connections
    # brine (MD hot side) loop
    m.fs.s01 = Arc(source=m.fs.feed.outlet, destination=m.fs.pump_feed.inlet)
    m.fs.s02 = Arc(source=m.fs.pump_feed.outlet, destination=m.fs.mixer.feed)
    m.fs.s03 = Arc(source=m.fs.mixer.outlet, destination=m.fs.hx.cold_inlet)
    m.fs.s04 = Arc(source=m.fs.hx.cold_outlet, destination=m.fs.pump_brine.inlet)
    m.fs.s05 = Arc(source=m.fs.pump_brine.outlet, destination=m.fs.heater.inlet)
    m.fs.s06 = Arc(source=m.fs.heater.outlet, destination=m.fs.MD.hot_ch_inlet)
    m.fs.s07 = Arc(
        source=m.fs.MD.hot_ch_outlet, destination=m.fs.separator_concentrate.inlet
    )
    m.fs.s08 = Arc(
        source=m.fs.separator_concentrate.reject, destination=m.fs.reject.inlet
    )
    m.fs.s09 = Arc(
        source=m.fs.separator_concentrate.recycle, destination=m.fs.mixer.recycle
    )

    # pemeate (MD cold side) loop
    m.fs.s10 = Arc(source=m.fs.chiller.outlet, destination=m.fs.MD.cold_ch_inlet)
    m.fs.s11 = Arc(source=m.fs.MD.cold_ch_outlet, destination=m.fs.hx.hot_inlet)
    m.fs.s12 = Arc(source=m.fs.hx.hot_outlet, destination=m.fs.separator_permeate.inlet)
    m.fs.s13 = Arc(
        source=m.fs.separator_permeate.permeate, destination=m.fs.permeate.inlet
    )
    m.fs.s14 = Arc(
        source=m.fs.separator_permeate.cold_loop_stream,
        destination=m.fs.pump_permeate.inlet,
    )
    m.fs.s15 = Arc(source=m.fs.pump_permeate.outlet, destination=m.fs.chiller.inlet)

    TransformationFactory("network.expand_arcs").apply_to(m)
    add_costs(m)

    # set default property values
    m.fs.properties_hot_ch.set_default_scaling(
        "flow_mass_phase_comp", 1, index=("Liq", "H2O")
    )
    m.fs.properties_hot_ch.set_default_scaling(
        "flow_mass_phase_comp", 1e2, index=("Liq", "TDS")
    )
    m.fs.properties_cold_ch.set_default_scaling(
        "flow_mass_phase_comp", 1, index=("Liq", "H2O")
    )
    m.fs.properties_cold_ch.set_default_scaling(
        "flow_mass_phase_comp", 1e2, index=("Liq", "TDS")
    )
    m.fs.properties_vapor.set_default_scaling(
        "flow_mass_phase_comp", 1, index=("Liq", "H2O")
    )
    m.fs.properties_vapor.set_default_scaling(
        "flow_mass_phase_comp", 1, index=("Vap", "H2O")
    )
    # set unit model values
    iscale.set_scaling_factor(m.fs.pump_permeate.control_volume.work, 1e-3)
    iscale.set_scaling_factor(m.fs.pump_brine.control_volume.work, 1e-3)
    iscale.set_scaling_factor(m.fs.pump_feed.control_volume.work, 1e-3)
    m.fs.feed.properties[0].flow_vol_phase["Liq"]
    m.fs.feed.properties[0].mass_frac_phase_comp["Liq", "TDS"]

    iscale.set_scaling_factor(m.fs.hx.hot.heat, 1e-3)
    iscale.set_scaling_factor(m.fs.hx.cold.heat, 1e-3)
    iscale.set_scaling_factor(m.fs.hx.overall_heat_transfer_coefficient, 1e-3)
    iscale.set_scaling_factor(m.fs.hx.area, 1e-1)
    iscale.set_scaling_factor(m.fs.MD.hot_ch.heat, 1e-3)
    iscale.set_scaling_factor(m.fs.MD.cold_ch.heat, 1e-3)
    iscale.set_scaling_factor(m.fs.heater.control_volume.heat, 1e-3)
    iscale.set_scaling_factor(m.fs.chiller.control_volume.heat, 1e-3)
    iscale.set_scaling_factor(m.fs.MD.area, 1e-1)

    iscale.calculate_scaling_factors(m)

    return m


def add_costs(m):
    m.fs.costing = WaterTAPCosting()
    m.fs.MD.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.heater.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_heater_chiller,
        costing_method_arguments={"HC_type": "electric_heater"},
    )
    m.fs.chiller.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_heater_chiller,
        costing_method_arguments={"HC_type": "chiller"},
    )
    m.fs.pump_feed.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_pump,
        costing_method_arguments={"pump_type": "low_pressure"},
    )
    m.fs.pump_permeate.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_pump,
        costing_method_arguments={"pump_type": "low_pressure"},
    )
    m.fs.pump_brine.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_pump,
        costing_method_arguments={"pump_type": "low_pressure"},
    )
    m.fs.hx.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.mixer.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    m.fs.costing.cost_process()
    m.fs.costing.add_annual_water_production(m.fs.permeate.properties[0].flow_vol)
    m.fs.costing.add_specific_energy_consumption(m.fs.permeate.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.permeate.properties[0].flow_vol)


def set_operating_conditions(m):
    # overall recovery
    m.fs.overall_recovery.fix(0.5)
    # feed
    feed_flow_mass = 1
    feed_mass_frac_TDS = 0.035
    feed_pressure = 101325  # atmospheric
    feed_temperature = 273.15 + 25
    feed_mass_frac_H2O = 1 - feed_mass_frac_TDS
    m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "TDS"].fix(
        feed_flow_mass * feed_mass_frac_TDS
    )
    m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )

    m.fs.feed.properties[0].pressure.fix(feed_pressure)  # Pa
    m.fs.feed.properties[0].temperature.fix(feed_temperature)  # K

    # concentrate separator
    # either the overall recovery or the split fraction will be fixed, but not both
    recycle_fraction = 0.9
    recycle_flow_mass = 10 * feed_flow_mass
    # MD
    membrane_pressure_drop = -5e5
    membrane_area = 10 * recycle_flow_mass
    m.fs.MD.area.fix(membrane_area)  # m^2
    m.fs.MD.permeability_coef.fix(1e-10)  # kg/m-s-Pa
    m.fs.MD.membrane_thickness.fix(1e-4)  # m
    m.fs.MD.membrane_thermal_conductivity.fix(0.2)  # W/K-m
    length = 3
    m.fs.MD.length.fix(length)  # m

    m.fs.MD.hot_ch.channel_height.fix(0.0019)  # m
    m.fs.MD.hot_ch.spacer_porosity.fix(0.77)
    m.fs.MD.cold_ch.channel_height.fix(0.0019)  # m
    m.fs.MD.cold_ch.spacer_porosity.fix(0.77)

    # heat exchanger
    heat_exchanger_area = 4 * membrane_area
    heat_exchanger_pressure_drop = -1e5
    m.fs.hx.overall_heat_transfer_coefficient.fix(2e3)  # W/K-m^2
    m.fs.hx.area.fix(heat_exchanger_area)  # m^2
    m.fs.hx.cold.deltaP[0].fix(heat_exchanger_pressure_drop)  # Pa
    m.fs.hx.hot.deltaP[0].fix(heat_exchanger_pressure_drop)  # Pa

    # feed pump
    m.fs.pump_feed.efficiency_pump.fix(0.8)
    m.fs.pump_feed.control_volume.deltaP[0].fix(7e5)  # Pa

    # brine pump
    LEP = 7e5
    heater_pressure_drop = -1e5
    m.fs.pump_brine.efficiency_pump.fix(0.8)
    m.fs.pump_brine.control_volume.properties_out[0].pressure.fix(
        feed_pressure + LEP + heater_pressure_drop + heat_exchanger_pressure_drop
    )

    # permeate pump
    chiller_pressure_drop = -1e5
    m.fs.pump_permeate.efficiency_pump.fix(0.8)
    m.fs.pump_permeate.control_volume.properties_out[0].pressure.fix(
        feed_pressure + LEP + chiller_pressure_drop + heat_exchanger_pressure_drop
    )

    # heater
    heater_top_temperature = 273.15 + 90
    mixed_brine_mass_frac_TDS = 0.07
    mixed_brine_mass_frac_H2O = 1 - mixed_brine_mass_frac_TDS
    mixed_brine_flow_mass = feed_flow_mass + recycle_flow_mass
    m.fs.heater.control_volume.properties_out[0].temperature.fix(
        heater_top_temperature
    )  # K
    m.fs.heater.control_volume.deltaP[0].fix(heater_pressure_drop)
    m.fs.heater.control_volume.properties_out[0].pressure.value = (
        m.fs.pump_brine.control_volume.properties_out[0].pressure.value
        - heater_pressure_drop
    )
    m.fs.heater.control_volume.properties_out[0].flow_mass_phase_comp[
        "Liq", "H2O"
    ].value = (mixed_brine_flow_mass * mixed_brine_mass_frac_H2O)
    m.fs.heater.control_volume.properties_out[0].flow_mass_phase_comp[
        "Liq", "TDS"
    ].value = (mixed_brine_flow_mass * mixed_brine_mass_frac_TDS)

    # chiller
    chiller_outlet_temperature = 273.15 + 10
    # The flow rate in MD cold and hot channels is assumed to be equal.
    cold_loop_stream_flow_mass = recycle_flow_mass + feed_flow_mass
    m.fs.chiller.control_volume.properties_out[0].temperature.fix(
        chiller_outlet_temperature
    )  # K
    m.fs.chiller.control_volume.deltaP[0].fix(chiller_pressure_drop)
    m.fs.chiller.control_volume.properties_out[0].pressure.value = (
        m.fs.pump_permeate.control_volume.properties_out[0].pressure.value
        - chiller_pressure_drop
    )

    m.fs.chiller.control_volume.properties_out[0].flow_mass_phase_comp[
        "Liq", "TDS"
    ].value = 0
    m.fs.chiller.control_volume.properties_out[0].flow_mass_phase_comp["Liq", "H2O"] = (
        m.fs.heater.control_volume.properties_out[0].flow_mass_phase_comp["Liq", "H2O"]
    )

    # check degrees of freedom
    if degrees_of_freedom(m) != 0:
        raise RuntimeError(
            "The set_operating_conditions function resulted in {} "
            "degrees of freedom rather than 0. This error suggests "
            "that too many or not enough variables are fixed for a "
            "simulation.".format(degrees_of_freedom(m))
        )


def solve(blk, solver=None, tee=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    if not check_optimal_termination(results):
        results = solver.solve(blk, tee=tee)
    return results


def initialize_system(m, solver=None, verbose=True):
    if solver is None:
        solver = get_solver()

    # initialize feed block
    m.fs.feed.initialize()

    # with fixed chiller and heater outlet temperature, and initial guess for the flowrates, MD is initialized
    propagate_state(m.fs.s06)
    propagate_state(m.fs.s10)
    m.fs.MD.initialize()

    # hot loop
    propagate_state(m.fs.s07)
    m.fs.separator_concentrate.initialize()

    propagate_state(m.fs.s08)
    m.fs.reject.initialize()

    propagate_state(m.fs.s01)
    m.fs.pump_feed.initialize()

    propagate_state(m.fs.s02)
    propagate_state(m.fs.s09)
    m.fs.mixer.initialize()
    m.fs.mixer.pressure_equality_constraints[0, 2].deactivate()

    propagate_state(m.fs.s03)
    propagate_state(m.fs.s11)
    m.fs.hx.initialize()

    propagate_state(m.fs.s04)
    m.fs.pump_brine.initialize()

    propagate_state(m.fs.s05)
    m.fs.heater.initialize()

    # cold loop
    propagate_state(m.fs.s12)
    m.fs.separator_permeate.initialize()

    propagate_state(m.fs.s13)
    m.fs.permeate.initialize()

    propagate_state(m.fs.s14)
    m.fs.pump_permeate.initialize()

    propagate_state(m.fs.s15)
    m.fs.chiller.initialize()

    propagate_state(m.fs.s06)
    propagate_state(m.fs.s10)
    m.fs.MD.initialize()

    print(f"DOF: {degrees_of_freedom(m)}")

    m.fs.costing.initialize()


def optimize_set_up(m):
    # add objective
    m.fs.objective = Objective(expr=m.fs.costing.LCOW)

    m.fs.MD.area.unfix()
    m.fs.MD.area.setlb(1)
    m.fs.MD.area.setub(150)

    m.fs.MD.length.unfix()
    m.fs.MD.length.setlb(0)
    m.fs.MD.length.setub(150)

    m.fs.hx.area.unfix()
    m.fs.hx.area.setlb(1)
    m.fs.hx.area.setub(750)

    m.fs.heater.control_volume.properties_out[0].temperature.unfix()
    m.fs.heater.control_volume.properties_out[0].temperature.setlb(273.15 + 10)
    m.fs.heater.control_volume.properties_out[0].temperature.setub(273.15 + 90)

    m.fs.chiller.control_volume.properties_out[0].temperature.unfix()
    m.fs.chiller.control_volume.properties_out[0].temperature.setlb(273.15 + 10)
    m.fs.chiller.control_volume.properties_out[0].temperature.setub(273.15 + 25)

    m.fs.pump_brine.control_volume.properties_out[0].pressure.unfix()
    m.fs.pump_brine.control_volume.properties_out[0].pressure.setlb(10e5)
    m.fs.pump_brine.control_volume.properties_out[0].pressure.setub(80e5)
    m.fs.pump_brine.deltaP.setlb(0)

    # additional constraints
    LEP = 10e5
    m.fs.eq_liquid_entry_pressure = Constraint(
        expr=m.fs.MD.hot_ch_inlet.pressure[0] <= LEP
    )

    assert_degrees_of_freedom(m, 6)


def optimize(m, solver=None):
    # --solve---
    return solve(m, solver=solver)


def display_system(m):
    print("---system metrics---")
    feed_flow_mass = sum(
        m.fs.feed.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "TDS"]
    )
    feed_mass_frac_TDS = (
        m.fs.feed.flow_mass_phase_comp[0, "Liq", "TDS"].value / feed_flow_mass
    )
    print("Feed: %.2f kg/s, %.0f ppm" % (feed_flow_mass, feed_mass_frac_TDS * 1e6))

    permeate_flow_mass = sum(
        m.fs.permeate.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "TDS"]
    )
    permeate_mass_frac_TDS = (
        m.fs.permeate.flow_mass_phase_comp[0, "Liq", "TDS"].value / permeate_flow_mass
    )
    print(
        "Permeate: %.3f kg/s, %.0f ppm"
        % (permeate_flow_mass, permeate_mass_frac_TDS * 1e6)
    )

    reject_flow_mass = sum(
        m.fs.reject.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "TDS"]
    )
    reject_mass_frac_TDS = (
        m.fs.reject.flow_mass_phase_comp[0, "Liq", "TDS"].value / reject_flow_mass
    )
    print(
        "Reject: %.3f kg/s, %.0f ppm" % (reject_flow_mass, reject_mass_frac_TDS * 1e6)
    )

    print("Overall recovery: %.1f%%" % (value(m.fs.overall_recovery) * 100))

    print("Recycle ratio: %.1f" % (value(m.fs.recycle_ratio[0])))

    print("Thermal efficiency: %.1f%%" % (value(m.fs.MD.thermal_efficiency[0]) * 100))

    print("effectiveness: %.1f%%" % (value(m.fs.MD.effectiveness[0]) * 100))

    print(
        "Energy Consumption: %.1f kWh/m3"
        % value(m.fs.costing.specific_energy_consumption)
    )
    print("Levelized cost of water: %.2f $/m3" % value(m.fs.costing.LCOW))


def display_design(m):
    print("---decision variables---")
    print("Membrane area %.1f m2" % (m.fs.MD.area.value))
    print("Heat exchanger area %.1f m2" % (m.fs.hx.area.value))
    print(
        "Heater\noutlet temperature: %.1f Celcius\npower %.2f kW"
        % (
            m.fs.heater.outlet.temperature[0].value - 273.15,
            m.fs.heater.heat_duty[0].value / 1e3,
        )
    )
    print(
        "Chiller\noutlet temperature: %.1f Celcius\npower %.2f kW"
        % (
            m.fs.chiller.outlet.temperature[0].value - 273.15,
            m.fs.chiller.heat_duty[0].value / 1e3,
        )
    )
    print(
        "brine pump\noutlet pressure: %.1f bar\npower %.2f kW"
        % (
            m.fs.pump_brine.outlet.pressure[0].value / 1e5,
            m.fs.pump_brine.work_mechanical[0].value / 1e3,
        )
    )
    print(
        "permeate pump\noutlet pressure: %.1f bar\npower %.2f kW"
        % (
            m.fs.pump_permeate.outlet.pressure[0].value / 1e5,
            m.fs.pump_permeate.work_mechanical[0].value / 1e3,
        )
    )


def display_state(m):
    print("---state---")

    def print_state(s, b):
        flow_mass = sum(
            b.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "TDS"]
        )
        mass_frac_ppm = b.flow_mass_phase_comp[0, "Liq", "TDS"].value / flow_mass * 1e6
        pressure_bar = b.pressure[0].value / 1e5
        Temperature_Celsius = b.temperature[0].value - 273.15
        print(
            s
            + ": %.3f kg/s, %.0f ppm, %.1f bar, %.1f °C"
            % (flow_mass, mass_frac_ppm, pressure_bar, Temperature_Celsius)
        )

    print_state("Feed      ", m.fs.feed.outlet)
    print_state("Feed pump out    ", m.fs.pump_feed.outlet)
    print_state("Mixer out    ", m.fs.mixer.outlet)
    print_state("Heat exchanger hot channel in", m.fs.hx.hot_side_inlet)
    print_state("Heat exchanger hot channel out", m.fs.hx.hot_side_outlet)
    print_state("Heat exchanger cold channel in", m.fs.hx.cold_side_inlet)
    print_state("Heat exchanger cold channel out", m.fs.hx.cold_side_outlet)
    print_state("brine pump out    ", m.fs.pump_brine.outlet)
    print_state("Heater out    ", m.fs.heater.outlet)
    print_state("MD hot channel in", m.fs.MD.hot_ch_inlet)
    print_state("MD hot channel out", m.fs.MD.hot_ch_outlet)
    print_state("MD cold channel in", m.fs.MD.cold_ch_inlet)
    print_state("MD cold channel out", m.fs.MD.cold_ch_outlet)
    print_state("Concentrate separator recycle ", m.fs.separator_concentrate.recycle)
    print_state("Concentrate separator reject ", m.fs.separator_concentrate.reject)
    print_state("permeate pump out    ", m.fs.pump_permeate.outlet)
    print_state("Concentrate permeate permeate ", m.fs.separator_permeate.permeate)
    print_state(
        "Concentrate permeate cold loop stream ",
        m.fs.separator_permeate.cold_loop_stream,
    )
    print_state("chiller out    ", m.fs.chiller.outlet)


if __name__ == "__main__":
    m = main()



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for full Water Resource Recovery Facility with ASM2d and ADM1
(WRRF; a.k.a., wastewater treatment plant) flowsheet example with ASM1 and ADM1.
The flowsheet follows the same formulation as benchmark simulation model no.2 (BSM2)
but comprises different specifications for default values than BSM2.
"""

# Some more information about this module
__author__ = "Chenyu Wang"

import pytest

from pyomo.environ import assert_optimal_termination, value
from pyomo.util.check_units import assert_units_consistent

from idaes.core.util.model_statistics import degrees_of_freedom

from watertap.flowsheets.anaerobic_digester.modified_ADM1_flowsheet_with_translators import (
    main,
)
from watertap.core.solvers import get_solver

solver = get_solver()


@pytest.mark.requires_idaes_solver
class TestADM1BioPFalse:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m, res = main(bio_P=False)
        m.results = res
        return m

    @pytest.mark.integration
    def test_structure(self, system_frame):
        assert_units_consistent(system_frame)
        assert degrees_of_freedom(system_frame) == 0
        assert_optimal_termination(system_frame.results)

    @pytest.mark.component
    def test_solve(self, system_frame):
        m = system_frame

        assert value(m.fs.Treated.properties[0].flow_vol) == pytest.approx(
            0.003, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_A"]) == pytest.approx(
            8.4935, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_F"]) == pytest.approx(
            22.163, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_I"]) == pytest.approx(
            0.05745, rel=1e-3
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_N2"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_NH4"]
        ) == pytest.approx(2.0167, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_NO3"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_O2"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_PO4"]
        ) == pytest.approx(65.479, rel=1e-3)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_K"]) == pytest.approx(
            1.0654, rel=1e-2
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_Mg"]
        ) == pytest.approx(0.71312, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_IC"]
        ) == pytest.approx(1.2023, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_AUT"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_H"]) == pytest.approx(
            0, abs=1e-6
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_I"]) == pytest.approx(
            14.875, rel=1e-3
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_PAO"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_PHA"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_PP"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_S"]) == pytest.approx(
            3.3158, rel=1e-3
        )


@pytest.mark.requires_idaes_solver
class TestADM1BioPTrue:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m, res = main(bio_P=True)
        m.results = res
        return m

    @pytest.mark.integration
    def test_structure(self, system_frame):
        assert_units_consistent(system_frame)
        assert degrees_of_freedom(system_frame) == 0
        assert_optimal_termination(system_frame.results)

    @pytest.mark.component
    def test_solve(self, system_frame):
        m = system_frame

        assert value(m.fs.Treated.properties[0].flow_vol) == pytest.approx(
            0.003, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_A"]) == pytest.approx(
            9.1528, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_F"]) == pytest.approx(
            23.908, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_I"]) == pytest.approx(
            0.05745, rel=1e-3
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_N2"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_NH4"]
        ) == pytest.approx(1.9658, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_NO3"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_O2"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_PO4"]
        ) == pytest.approx(4.0233, rel=1e-3)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_K"]) == pytest.approx(
            1.2776, rel=1e-2
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_Mg"]
        ) == pytest.approx(0.92538, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_IC"]
        ) == pytest.approx(0.63549, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_AUT"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_H"]) == pytest.approx(
            0, abs=1e-6
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_I"]) == pytest.approx(
            14.881, rel=1e-3
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_PAO"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_PHA"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_PP"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_S"]) == pytest.approx(
            0.85963, rel=1e-3
        )



from pyomo.environ import assert_optimal_termination, value
from watertap.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom

from watertap.flowsheets.generic_desalination_train import generic_train as gt

import pytest

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.mark.unit
def test_standard_train():
    # This will test bilding all the units and
    # unit operations as well as opitmization
    # TODO: Might want to separate tests based on unit, and refine but probably not
    # needed as this is a very simple model
    m = gt.build()
    m.fs.feed.properties[0].temperature.fix(298.15)
    m.fs.feed.properties[0].pressure.fix(101325)
    m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"].fix(1)
    m.fs.feed.properties[0].conc_mass_phase_comp["Liq", "TDS"].fix(3.5)  # kg/m3 or g/L
    m.fs.feed.properties[0].conc_mass_phase_comp["Liq", "X"].fix(0.3)  # kg/m3 or g/L
    gt.update_feed(m.fs)

    gt.initialize(m)
    m.fs.Pretreatment.separator.component_removal_percent["X"].fix(50)
    m.fs.Pretreatment.separator.separation_cost["X"].fix(0.5)

    m.fs.Desal_1.desalter.water_recovery.fix(80)
    m.fs.Desal_2.desalter.water_recovery.fix(50)
    m.fs.Desal_2.desalter.recovery_cost.fix(0.01)
    m.fs.Desal_2.desalter.recovery_cost_offset.fix(35)
    # config as crystalizer
    m.fs.Desal_3.desalter.water_recovery.unfix()
    m.fs.Desal_3.desalter.brine_water_mass_percent.fix(80)

    m.fs.Valorizer.separator.product_value["X"].fix(1)
    m.fs.Valorizer.separator.component_removal_percent["X"].fix(50)

    result = gt.solve(m)
    assert_optimal_termination(result)
    assert degrees_of_freedom(m) == 0
    # test overall result costs operation
    assert value(m.fs.water_recovery) == pytest.approx(98.220, rel=1e-3)

    # test costing compute
    assert value(m.fs.costing.LCOW) == pytest.approx(1.2686, rel=1e-3)
    unit_costs = {
        "Product distribution": 0,
        "Waste disposal": 0.0180462225724005,
        "Feed sourcing": 0,
        "Pretreatment": 0.17817144384075734,
        "Desal_1": 0.24431276721281453,
        "Desal_2": 0.06616804112013724,
        "Desal_3": 0.8382712295193712,
        "Valorizer cost": 0,
        "Valorizer revenue": -0.07635918566495165,
    }
    for name, var in m.fs.costing.LCOW_unit.items():
        assert value(var) == pytest.approx(unit_costs[name], rel=1e-5)

    assert value(
        m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"]
    ) == pytest.approx(1, rel=1e-3)
    assert value(
        m.fs.product.properties[0].flow_mass_phase_comp["Liq", "H2O"]
    ) == pytest.approx(0.9859465972001905, rel=1e-3)
    assert value(
        m.fs.disposal.properties[0].flow_mass_phase_comp["Liq", "H2O"]
    ) == pytest.approx(0.014053402799809303, rel=1e-5)

    # test desalter brine water content constraint
    assert value(m.fs.Desal_3.desalter.water_recovery) == pytest.approx(
        82.34735228114, rel=1e-3
    )

    # test pretreatment removal constraints
    assert value(
        m.fs.Pretreatment.separator.product_properties[0].flow_mass_phase_comp[
            "Liq", "X"
        ]
    ) == pytest.approx(0.00015057215854267565, rel=1e-6)

    # test valorizer removal constraints
    assert value(
        m.fs.Valorizer.separator.product_properties[0].flow_mass_phase_comp["Liq", "X"]
    ) == pytest.approx(7.528607927133832e-05, rel=1e-6)


@pytest.mark.unit
def test_main():
    result = gt.main()
    assert_optimal_termination(result)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import units as pyunits
from pyomo.util.check_units import assert_units_equivalent
from watertap.tools.oli_api.util.fixed_keys_dict import output_unit_set, input_unit_set


@pytest.mark.unit
def test_fixed_keys_dict():
    with pytest.raises(RuntimeError):
        output_unit_set["invalid_key"] = "value"

    with pytest.raises(Exception):
        del output_unit_set["any_key"]

    with pytest.raises(RuntimeError):
        output_unit_set._check_value("mass", ["not_kg"])


input_unit_set_temp = {
    "inflows": {
        "oli_unit": "mg/L",
        "pyomo_unit": pyunits.mg / pyunits.L,
    },
    "molecularConcentration": {
        "oli_unit": "mg/L",
        "pyomo_unit": pyunits.mg / pyunits.L,
    },
    "mass": {"oli_unit": "mg", "pyomo_unit": pyunits.mg},
    "temperature": {"oli_unit": "K", "pyomo_unit": pyunits.K},
    "pressure": {"oli_unit": "Pa", "pyomo_unit": pyunits.Pa},
    "enthalpy": {"oli_unit": "J", "pyomo_unit": pyunits.J},
    "vaporAmountMoles": {"oli_unit": "mol", "pyomo_unit": pyunits.mol},
    "vaporMolFrac": {
        "oli_unit": "mol/mol",
        "pyomo_unit": pyunits.mol / pyunits.mol,
    },
    "totalVolume": {"oli_unit": "L", "pyomo_unit": pyunits.L},
    "pipeDiameter": {"oli_unit": "m", "pyomo_unit": pyunits.meter},
    "pipeFlowVelocity": {
        "oli_unit": "m/s",
        "pyomo_unit": pyunits.meter / pyunits.second,
    },
    "diskDiameter": {"oli_unit": "m", "pyomo_unit": pyunits.meter},
    "diskRotatingSpeed": {"oli_unit": "cycle/s", "pyomo_unit": 1 / pyunits.second},
    "rotorDiameter": {"oli_unit": "m", "pyomo_unit": pyunits.meter},
    "rotorRotation": {"oli_unit": "cycle/s", "pyomo_unit": 1 / pyunits.second},
    "shearStress": {"oli_unit": "Pa", "pyomo_unit": pyunits.Pa},
    "pipeDiameter": {"oli_unit": "m", "pyomo_unit": pyunits.meter},
    "pipeRoughness": {"oli_unit": "m", "pyomo_unit": pyunits.meter},
    "liquidFlowInPipe": {
        "oli_unit": "L/s",
        "pyomo_unit": pyunits.L / pyunits.second,
    },
    "gasFlowInPipe": {"oli_unit": "L/s", "pyomo_unit": pyunits.L / pyunits.second},
    "viscAbs2ndLiq": {
        "oli_unit": "Pa-s",
        "pyomo_unit": pyunits.Pa * pyunits.second,
    },
    "alkalinity": {"oli_unit": "mg HCO3/L", "pyomo_unit": pyunits.mg / pyunits.L},
    "TIC": {"oli_unit": "mol C/L", "pyomo_unit": pyunits.mol / pyunits.L},
    "CO2GasFraction": {
        "oli_unit": "mol/mol",
        "pyomo_unit": pyunits.mol / pyunits.mol,
    },
}

output_unit_set_temp = {
    "enthalpy": input_unit_set_temp["enthalpy"],
    "mass": input_unit_set_temp["mass"],
    "pt": input_unit_set_temp["pressure"],
    "total": input_unit_set_temp["mass"],
    "liq1_phs_comp": input_unit_set_temp["mass"],
    "solid_phs_comp": input_unit_set_temp["mass"],
    "vapor_phs_comp": input_unit_set_temp["mass"],
    "liq2_phs_comp": input_unit_set_temp["mass"],
    "combined_phs_comp": input_unit_set_temp["mass"],
    "molecularConcentration": input_unit_set_temp["molecularConcentration"],
}


@pytest.mark.unit
def test_input_output_fixed_key_dicts():
    input_unit_set_temp.keys() == input_unit_set.keys()
    for k in input_unit_set_temp.keys():
        assert input_unit_set[k]["oli_unit"] == input_unit_set_temp[k]["oli_unit"]
        assert_units_equivalent(
            input_unit_set[k]["pyomo_unit"], input_unit_set_temp[k]["pyomo_unit"]
        )

    output_unit_set_temp.keys() == output_unit_set.keys()
    for k in output_unit_set_temp.keys():
        assert output_unit_set[k]["oli_unit"] == output_unit_set_temp[k]["oli_unit"]
        assert_units_equivalent(
            output_unit_set[k]["pyomo_unit"], output_unit_set_temp[k]["pyomo_unit"]
        )


@pytest.mark.unit
def test_input_unit_set():
    unit_set = input_unit_set
    # check defaults for one of the properties
    assert unit_set["molecularConcentration"]["oli_unit"] == "mg/L"
    assert str(unit_set["molecularConcentration"]["pyomo_unit"]) == "mg/L".lower()
    assert_units_equivalent(
        (unit_set["molecularConcentration"]["pyomo_unit"]), pyunits.mg / pyunits.L
    )
    assert hasattr(
        (unit_set["molecularConcentration"]["pyomo_unit"]), "is_expression_type"
    )
    assert str(unit_set["molecularConcentration"]["pyomo_unit"]) == "mg/L".lower()

    # reset oli_unit with a different unit, provided as string, and check that pyomo_unit follows along
    unit_set["molecularConcentration"]["oli_unit"] = "mol/L"
    assert unit_set["molecularConcentration"]["oli_unit"] == "mol/L"
    assert str(unit_set["molecularConcentration"]["pyomo_unit"]) == "mol/L".lower()
    assert_units_equivalent(
        (unit_set["molecularConcentration"]["pyomo_unit"]), pyunits.mol / pyunits.L
    )

    # reset pyomo_unit with a different unit, provided as pint units, and check that oli_unit follows along
    unit_set["molecularConcentration"]["pyomo_unit"] = pyunits.mg
    assert_units_equivalent(
        (unit_set["molecularConcentration"]["pyomo_unit"]), pyunits.mg
    )
    assert unit_set["molecularConcentration"]["oli_unit"] == "mg"

    # reset pyomo_unit with a different unit, provided as string units, and check that oli_unit follows along
    unit_set["molecularConcentration"]["pyomo_unit"] = "mg/L"
    assert_units_equivalent(
        (unit_set["molecularConcentration"]["pyomo_unit"]), pyunits.mg / pyunits.L
    )
    assert unit_set["molecularConcentration"]["oli_unit"] == "mg/L"

    # reset oli_unit with a different unit, provided as pint units, and check that pyomo_unit follows along
    unit_set["molecularConcentration"]["oli_unit"] = pyunits.mol / pyunits.L
    assert unit_set["molecularConcentration"]["oli_unit"] == "mol/L".lower()
    assert str(unit_set["molecularConcentration"]["pyomo_unit"]) == "mol/L".lower()
    assert_units_equivalent(
        (unit_set["molecularConcentration"]["pyomo_unit"]), pyunits.mol / pyunits.L
    )

    with pytest.raises(
        RuntimeError,
        match="Setting 1 as the value for oli_unit is not permitted as a value for oli and pyomo units. Please enter units as a string type or pint units.",
    ):
        unit_set["molecularConcentration"]["oli_unit"] = 1

    with pytest.raises(
        RuntimeError,
        match="Setting 1 as the value for pyomo_unit is not permitted as a value for oli and pyomo units. Please enter units as a string type or pint units.",
    ):
        unit_set["molecularConcentration"]["pyomo_unit"] = 1

    with pytest.raises(
        RuntimeError,
        match="Setting 1 as the value for molecularConcentration is not permitted as a value for oli and pyomo units. Please enter units as a string type or pint units.",
    ):
        unit_set["molecularConcentration"] = 1



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import (
    value,
    assert_optimal_termination,
)
from watertap.core.solvers import get_solver
from watertap.core.util.initialization import assert_degrees_of_freedom
from pyomo.util.check_units import assert_units_consistent
from watertap.flowsheets.dye_desalination.dye_desalination import (
    build,
    set_operating_conditions,
    initialize_system,
    initialize_costing,
    solve,
    add_costing,
    optimize_operation,
    display_results,
    display_costing,
)

solver = get_solver()


class TestDyewithROFlowsheetwithPretreatment:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build(include_pretreatment=True)
        return m

    @pytest.mark.unit
    def test_build(self, system_frame):
        m = system_frame
        assert_degrees_of_freedom(m, 28)
        assert_units_consistent(m)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame
        set_operating_conditions(m)
        initialize_system(m)

        # test feed
        assert pytest.approx(77.607, rel=1e-3) == value(
            m.fs.feed.flow_mass_comp[0, "H2O"]
        )

        assert pytest.approx(2, rel=1e-5) == value(m.fs.feed.conc_mass_comp[0, "tds"])

        assert pytest.approx(0.2, rel=1e-5) == value(m.fs.feed.conc_mass_comp[0, "dye"])

        # test wwtp
        assert pytest.approx(0.14, rel=1e-5) == value(
            m.fs.pretreatment.wwtp.energy_electric_flow_vol_inlet
        )

        # test pump block
        assert pytest.approx(7, rel=1e-5) == value(
            m.fs.dye_separation.P1.applied_pressure[0]
        )

        # test nanofiltration
        assert pytest.approx(359.7653, rel=1e-5) == value(
            m.fs.dye_separation.nanofiltration.area
        )

        # check products
        assert pytest.approx(0, rel=1e-6) == value(
            m.fs.wwt_retentate.flow_mass_comp[0, "dye"]
        )
        assert pytest.approx(0.018391, rel=1e-3) == value(
            m.fs.concentrated_dye.flow_mass_comp[0, "tds"]
        )

        assert pytest.approx(0.965, rel=1e-5) == value(
            m.fs.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(0.035, rel=1e-5) == value(
            m.fs.brine.flow_mass_phase_comp[0, "Liq", "TDS"]
        )

    @pytest.mark.component
    def test_solve(self, system_frame):
        m = system_frame

        results = solve(m)

        # check products
        assert pytest.approx(0.02894, rel=1e-3) == value(
            m.fs.wwt_retentate.flow_mass_comp[0, "tds"]
        )
        assert pytest.approx(7.760, rel=1e-3) == value(
            m.fs.concentrated_dye.flow_mass_comp[0, "H2O"]
        )

        assert pytest.approx(34.93575, rel=1e-5) == value(
            m.fs.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(34.9102, rel=1e-5) == value(
            m.fs.brine.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

    @pytest.mark.component
    def test_costing(self, system_frame):
        m = system_frame

        add_costing(m)
        initialize_costing(m)
        optimize_operation(m)

        results = solve(m)
        assert_optimal_termination(results)

        # check costing
        assert pytest.approx(19.824344, rel=1e-3) == value(m.fs.LCOW)
        assert pytest.approx(15.159548, rel=1e-3) == value(m.fs.LCOT)

    @pytest.mark.component
    def test_display(self, system_frame):
        m = system_frame
        display_results(m)
        display_costing(m)


class TestDyewithROFlowsheetDefault:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build()
        return m

    @pytest.mark.unit
    def test_build(self, system_frame):
        m = system_frame
        assert_degrees_of_freedom(m, 24)
        assert_units_consistent(m)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame
        set_operating_conditions(m)
        initialize_system(m)

        # test feed
        assert pytest.approx(77.607, rel=1e-3) == value(
            m.fs.feed.flow_mass_comp[0, "H2O"]
        )

        assert pytest.approx(2, rel=1e-5) == value(m.fs.feed.conc_mass_comp[0, "tds"])

        assert pytest.approx(0.2, rel=1e-5) == value(m.fs.feed.conc_mass_comp[0, "dye"])

        # test pump block
        assert pytest.approx(7, rel=1e-5) == value(
            m.fs.dye_separation.P1.applied_pressure[0]
        )

        # test nanofiltration
        assert pytest.approx(359.8925, rel=1e-5) == value(
            m.fs.dye_separation.nanofiltration.area
        )

        # check products
        assert pytest.approx(0.022595, rel=1e-3) == value(
            m.fs.concentrated_dye.flow_mass_comp[0, "tds"]
        )

        assert pytest.approx(0.965, rel=1e-5) == value(
            m.fs.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(0.035, rel=1e-5) == value(
            m.fs.brine.flow_mass_phase_comp[0, "Liq", "TDS"]
        )

    @pytest.mark.component
    def test_solve(self, system_frame):
        m = system_frame

        results = solve(m)

        # check products
        assert pytest.approx(7.760, rel=1e-3) == value(
            m.fs.concentrated_dye.flow_mass_comp[0, "H2O"]
        )

        assert pytest.approx(34.9387, rel=1e-5) == value(
            m.fs.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(34.9073, rel=1e-5) == value(
            m.fs.brine.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(1295.2263, rel=1e-5) == value(m.fs.desalination.RO.area)

        assert pytest.approx(-41307.369, rel=1e-5) == value(
            m.fs.desalination.RO.deltaP[0]
        )

        assert pytest.approx(3.2693848e-7, rel=1e-5) == value(
            m.fs.desalination.RO.flux_mass_phase_comp[0, 1, "Liq", "TDS"]
        )

    def test_costing(self, system_frame):
        m = system_frame

        add_costing(m)
        initialize_costing(m)
        optimize_operation(m)

        results = solve(m)
        assert_optimal_termination(results)

        # check costing
        assert pytest.approx(19.7323247, rel=1e-3) == value(m.fs.LCOW)
        assert pytest.approx(15.0766, rel=1e-3) == value(m.fs.LCOT)

    @pytest.mark.component
    def test_display(self, system_frame):
        m = system_frame
        display_results(m)
        display_costing(m)


class TestDyewith0DROFlowsheet:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build(RO_1D=False)
        return m

    @pytest.mark.unit
    def test_build(self, system_frame):
        m = system_frame
        assert_degrees_of_freedom(m, 24)
        assert_units_consistent(m)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame
        set_operating_conditions(m)
        initialize_system(m)

        # test feed
        assert pytest.approx(77.607, rel=1e-3) == value(
            m.fs.feed.flow_mass_comp[0, "H2O"]
        )

        assert pytest.approx(2, rel=1e-5) == value(m.fs.feed.conc_mass_comp[0, "tds"])

        assert pytest.approx(0.2, rel=1e-5) == value(m.fs.feed.conc_mass_comp[0, "dye"])

        # test pump block
        assert pytest.approx(7, rel=1e-5) == value(
            m.fs.dye_separation.P1.applied_pressure[0]
        )

        # test nanofiltration
        assert pytest.approx(359.8925, rel=1e-5) == value(
            m.fs.dye_separation.nanofiltration.area
        )

        # check products
        assert pytest.approx(0.022595, rel=1e-3) == value(
            m.fs.concentrated_dye.flow_mass_comp[0, "tds"]
        )

        assert pytest.approx(0.965, rel=1e-5) == value(
            m.fs.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(0.035, rel=1e-5) == value(
            m.fs.brine.flow_mass_phase_comp[0, "Liq", "TDS"]
        )

    @pytest.mark.component
    def test_solve(self, system_frame):
        m = system_frame

        results = solve(m)

        # check products
        assert pytest.approx(7.760, rel=1e-3) == value(
            m.fs.concentrated_dye.flow_mass_comp[0, "H2O"]
        )

        assert pytest.approx(34.9387, rel=1e-5) == value(
            m.fs.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(34.9073, rel=1e-5) == value(
            m.fs.brine.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(-45144.537, rel=1e-5) == value(
            m.fs.desalination.RO.deltaP[0]
        )

        assert pytest.approx(1299.7286, rel=1e-5) == value(m.fs.desalination.RO.area)

        assert pytest.approx(3.26732341e-7, rel=1e-5) == value(
            m.fs.desalination.RO.flux_mass_phase_comp[0, 1, "Liq", "TDS"]
        )

    def test_costing(self, system_frame):
        m = system_frame

        add_costing(m)
        initialize_costing(m)
        optimize_operation(m)

        results = solve(m)
        assert_optimal_termination(results)

        # check costing
        assert pytest.approx(20.379757, rel=1e-3) == value(m.fs.LCOW)
        assert pytest.approx(15.154258, rel=1e-3) == value(m.fs.LCOT)

    @pytest.mark.component
    def test_display(self, system_frame):
        m = system_frame
        display_results(m)
        display_costing(m)


class TestDyewithROFlowsheetwithDewatering:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build(include_dewatering=True)
        return m

    @pytest.mark.unit
    def test_build(self, system_frame):
        m = system_frame
        assert_degrees_of_freedom(m, 27)
        assert_units_consistent(m)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame
        set_operating_conditions(m)
        initialize_system(m)

        # test feed
        assert pytest.approx(77.607, rel=1e-3) == value(
            m.fs.feed.flow_mass_comp[0, "H2O"]
        )

        assert pytest.approx(2, rel=1e-5) == value(m.fs.feed.conc_mass_comp[0, "tds"])

        assert pytest.approx(0.2, rel=1e-5) == value(m.fs.feed.conc_mass_comp[0, "dye"])

        # test dewaterer block
        assert pytest.approx(0.99, rel=1e-5) == value(
            m.fs.dewaterer.split_fraction[0, "precipitant", "dye"]
        )

        # test pump block
        assert pytest.approx(7, rel=1e-5) == value(
            m.fs.dye_separation.P1.applied_pressure[0]
        )

        # test nanofiltration
        assert pytest.approx(359.8925, rel=1e-5) == value(
            m.fs.dye_separation.nanofiltration.area
        )

        # check products
        assert pytest.approx(1, rel=1e-3) == value(
            m.fs.concentrated_dye.flow_mass_comp[0, "tds"]
        )

        assert pytest.approx(0.965, rel=1e-5) == value(
            m.fs.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(0.035, rel=1e-5) == value(
            m.fs.brine.flow_mass_phase_comp[0, "Liq", "TDS"]
        )

    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_solve(self, system_frame):
        m = system_frame

        results = solve(m)

        # check products
        assert pytest.approx(7.68306, rel=1e-3) == value(
            m.fs.treated.flow_mass_comp[0, "H2O"]
        )

        assert pytest.approx(0.077607, rel=1e-3) == value(
            m.fs.concentrated_dye.flow_mass_comp[0, "H2O"]
        )

        assert pytest.approx(34.9387, rel=1e-5) == value(
            m.fs.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(34.9073, rel=1e-5) == value(
            m.fs.brine.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_costing(self, system_frame):
        m = system_frame

        add_costing(m)
        initialize_costing(m)
        optimize_operation(m)

        results = solve(m)
        assert_optimal_termination(results)

        # check costing
        assert pytest.approx(1.057, rel=1e-3) == value(m.fs.LCOW)
        assert pytest.approx(-0.57517, rel=1e-3) == value(m.fs.LCOT)

    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_display(self, system_frame):
        m = system_frame
        display_results(m)
        display_costing(m)


class TestDyewithROFlowsheetwithGAC:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build(include_gac=True)
        return m

    @pytest.mark.unit
    def test_build(self, system_frame):
        m = system_frame
        assert_degrees_of_freedom(m, 43)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame
        set_operating_conditions(m)
        assert_units_consistent(m)
        initialize_system(m)

        # test feed
        assert pytest.approx(77.607, rel=1e-3) == value(
            m.fs.feed.flow_mass_comp[0, "H2O"]
        )

        assert pytest.approx(2, rel=1e-5) == value(m.fs.feed.conc_mass_comp[0, "tds"])

        assert pytest.approx(0.2, rel=1e-5) == value(m.fs.feed.conc_mass_comp[0, "dye"])

        # test pump block
        assert pytest.approx(7, rel=1e-5) == value(
            m.fs.dye_separation.P1.applied_pressure[0]
        )

        # test nanofiltration
        assert pytest.approx(359.8925, rel=1e-5) == value(
            m.fs.dye_separation.nanofiltration.area
        )

    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_solve(self, system_frame):
        m = system_frame

        results = solve(m)

        # check products
        assert pytest.approx(34.938678, rel=1e-6) == value(
            m.fs.permeate.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(7.7607, rel=1e-3) == value(
            m.fs.treated.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(0.0005, abs=1e-3) == value(
            m.fs.treated.flow_mass_phase_comp[0, "Liq", "dye"]
        )

        assert pytest.approx(0.022595, rel=1e-3) == value(
            m.fs.treated.flow_mass_phase_comp[0, "Liq", "tds"]
        )

        assert pytest.approx(0, rel=1e-3) == value(
            m.fs.concentrated_dye.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        assert pytest.approx(0.0149, abs=1e-3) == value(
            m.fs.concentrated_dye.flow_mass_phase_comp[0, "Liq", "dye"]
        )

        assert pytest.approx(0, rel=1e-3) == value(
            m.fs.concentrated_dye.flow_mass_phase_comp[0, "Liq", "tds"]
        )

    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_costing(self, system_frame):
        m = system_frame

        add_costing(m)
        initialize_costing(m)
        optimize_operation(m)

        results = solve(m)
        assert_optimal_termination(results)

        # check costing
        assert pytest.approx(0.4781649, rel=1e-3) == value(m.fs.LCOW)
        assert pytest.approx(-1.116956, rel=1e-3) == value(m.fs.LCOT)

    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_display(self, system_frame):
        m = system_frame
        display_results(m)
        display_costing(m)
